# we target 3.8+, so this should be okay without fallback to importlib_metadata
import importlib.metadata

# single source of truth for package version,
# see https://packaging.python.org/en/latest/single_source_version/

__version__ = importlib.metadata.version('colmena')



"""Tests across different queue implementations"""
from multiprocessing import Pool, Process
import pickle as pkl

from pytest import fixture, raises, mark
from redis import Redis, ConnectionError

from colmena.exceptions import TimeoutException, KillSignalException
from colmena.queue.base import ColmenaQueues
from colmena.queue.python import PipeQueues
from colmena.queue.redis import RedisQueues

# Determine which classes to test

has_redis = True
try:
    client = Redis()
    client.ping()
except ConnectionError:
    has_redis = False

to_test = [PipeQueues]
if has_redis:
    to_test.append(RedisQueues)


@fixture(params=to_test)
def queue(request) -> ColmenaQueues:
    return request.param(['a', 'b'])


def test_topics(queue):
    assert queue.topics == {'default', 'a', 'b'}


@mark.timeout(5)
def test_flush(queue):
    # Test that it flushes out the input queue
    queue.send_inputs(1, method='method')
    queue.flush()

    with raises(TimeoutException):
        queue.get_task(timeout=0.1)

    # Test that it will flush a result
    queue.send_inputs(1, method='method')
    topic, result = queue.get_task()
    queue.send_result(result)
    queue.flush()

    with raises(TimeoutException):
        queue.get_result(topic, timeout=0.1)


@mark.parametrize('topic', ['default', 'a'])
@mark.timeout(5)
def test_basic(queue, topic):
    """Make sure topics get passed back-and-forth correctly"""

    # Send a result
    with Pool(1) as pool:
        queue.send_inputs(1, method='method', topic=topic)
        my_topic, result = pool.apply(queue.get_task)
        assert my_topic == topic

        # Send it back
        result.deserialize()
        result.set_result(1, 1)
        result.serialize()
        pool.apply(queue.send_result, (result,))

        # Make sure it does not appear in b
        with raises(TimeoutException):
            queue.get_result(timeout=0.1, topic='b')
        queue.get_result(topic=topic)


@mark.timeout(5)
def test_kill_signal(queue):
    queue.send_kill_signal()
    with raises(KillSignalException):
        queue.get_task()


@mark.timeout(5)
def test_pickle(queue):
    """Ensure that we can still send and receive after pickling"""
    serialized = pkl.dumps(queue)

    # Deserializing the Connection objects owned by PipeQueues must be done
    # in a different process so that two Connection objects referencing the
    # same file descriptor do not exist in the same process. Otherwise, the
    # file descriptor will be closed twice when the objects are garbage
    # collected, causing an error.
    def _check(serialized: bytes) -> None:
        queue: ColmenaQueues = pkl.loads(serialized)
        queue.send_inputs(1, method='test')
        _, task = queue.get_task()
        assert task.method == 'test'

    process = Process(target=_check, args=(serialized,))
    process.start()
    process.join()
    assert process.exitcode == 0



"""Base classes for the Task Server and associated functions"""
import logging
from inspect import isgeneratorfunction
from abc import ABCMeta, abstractmethod
from concurrent.futures import Future
from multiprocessing import Process
from typing import Collection, Optional, Callable, Union

from colmena.exceptions import KillSignalException, TimeoutException
from colmena.models.methods import ColmenaMethod, PythonGeneratorMethod, PythonMethod
from colmena.models import Result, FailureInformation
from colmena.queue.base import ColmenaQueues

logger = logging.getLogger(__name__)


class BaseTaskServer(Process, metaclass=ABCMeta):
    """Abstract class for the Colmena Task Server, which manages the execution of tasks

    Start the task server by first instantiating it and then calling :meth:`start` to launch the server in a separate process.
    Clients submit task requests to the server by pushing them to a Redis queue, and then receive results from a second queue.

    The task server can be stopped by pushing a ``None`` to the task queue, signaling that no new tasks will be incoming.
    The remaining tasks will continue to be pushed to the output queue.

    ## Implementing a Task Server

    Different implementations vary in how the queue is processed.

    Each implementation must provide the :meth:`process_queue` function is responsible for executing tasks supplied on the tasks queue
    and ensuring completed results are written back to the result queue on completion.
    Tasks must first be wrapped in the :meth:`run_and_record_timing` decorator function to capture the runtime information.

    Implementations should also provide a `_cleanup` function that releases any resources reserved by the task server.
    """

    def __init__(self, queues: ColmenaQueues, method_names: Collection[str], timeout: Optional[int] = None):
        """
        Args:
            queues (TaskServerQueues): Queues for the task server
            timeout (int): Timeout for reading from the task queue, if desired
        """
        super().__init__()
        self.queues = queues
        self.timeout = timeout
        self.method_names = set(method_names)

    @abstractmethod
    def process_queue(self, topic: str, task: Result):
        """Execute a single task from the task queue

        Args:
            topic: Which task queue this result came from
            task: Task description
        """
        pass

    def listen_and_launch(self):
        logger.info('Begin pulling from task queue')
        while True:
            try:
                # Get a result from the queue
                topic, task = self.queues.get_task(self.timeout)
                logger.info(f'Received request for {task.method} with topic {topic}')

                # Make sure the method name is valid
                if task.method in self.method_names:
                    # Provide it to the workflow system to be executed
                    self.process_queue(topic, task)
                else:
                    task.success = False
                    task.failure_info = FailureInformation.from_exception(
                        ValueError(f'Method name "{task.method}" not recognized. Options: {", ".join(self.method_names)}')
                    )
                    self.queues.send_result(task)

            except KillSignalException:
                logger.info('Kill signal received')
                return
            except TimeoutException:
                logger.info('Timeout while waiting on task queue')
                return

    def _cleanup(self):
        """Close out any resources needed by the task server"""
        pass

    def _setup(self):
        """Start any resources needed by the task server after it has started in a new process"""
        pass

    def run(self) -> None:
        """Launch the thread and start running tasks

        Blocks until the inputs queue is closed and all tasks have completed"""
        logger.info(f"Started task server {self.__class__.__name__} on {self.ident}")
        self.queues.set_role('server')

        # Perform any setup operations
        self._setup()

        # Loop until queue has closed
        self.listen_and_launch()

        # Shutdown any needed functions
        self._cleanup()


class FutureBasedTaskServer(BaseTaskServer, metaclass=ABCMeta):
    """Base class for workflow engines that use Python's native Future object

    Implementations need to specify a function, :meth:`_submit`, that creates the Future and
    `FutureBasedTaskServer`'s implementation of :meth:`process_queue` will add a
    callback to submit the output to the result queue.
    Note that implementations are still responsible for adding the :meth:`run_and_record_timing` decorator.
    """

    def perform_callback(self, future: Future, result: Result, topic: str):
        """Send a completed result back to queue. Used as a callback for complete tasks

        Args:
            future: Future for a task
            result: Initial result object. Used if the future throws an exception
            topic: Topic used to send back to the user
        """

        # Get any exception thrown by the workflow engine
        task_exc = future.exception()

        # If it was, send back a modified copy of the input structure
        if task_exc is not None:
            # Mark it as unsuccessful and capture the exception information
            result.success = False
            result.failure_info = FailureInformation.from_exception(task_exc)
        else:
            # If not, the result object is the one we need
            result = future.result()

        # If the task failed and we have retries left, try again
        if not result.success and result.retries < result.max_retries:
            # Increment the retry count and clear the failure information
            result.retries += 1
            result.failure_info, result.success = None, None

            # Log the retry
            logger.warning(f'Task {result.task_id} failed. Retrying with {result.retries}/{result.max_retries} retries.')

            # Provide it to the workflow system to be re-executed
            self.process_queue(topic, result)

            # Do not send the result back to the user
            return

        # Mark the task received timestamp
        result.mark_task_received()

        # Put them back in the pipe with the proper topic
        self.queues.send_result(result)

    @abstractmethod
    def _submit(self, task: Result, topic: str) -> Optional[Future]:
        """Submit the task to the workflow engine

        Args:
            task: Task description
            topic: Topic for the task
        Returns:
            Future for the result object, if any that needs a "return to user" callback is created
        """
        pass

    def process_queue(self, topic: str, task: Result):
        # Launch the task
        future = self._submit(task, topic)

        # Create the callback
        if future is not None:
            future.add_done_callback(lambda x: self.perform_callback(x, task, topic))


def convert_to_colmena_method(function: Union[Callable, ColmenaMethod]) -> ColmenaMethod:
    """Wrap user-supplified functions in the task model wrapper, if needed

    Args:
        function: User-provided function
    Returns:
        Function as appropriate subclasses of Colmena Task wrapper
    """

    name = function.__name__
    if isinstance(function, ColmenaMethod):
        return function
    elif isgeneratorfunction(function):
        logger.info(f'Assigned a PythonGeneratorMethod wrapper to {name}')
        return PythonGeneratorMethod(function)
    else:
        logger.info(f'Assigned a PythonMethod wrapper to {name}')
        return PythonMethod(function)



"""Tests for the Parsl implementation of the task server"""
from typing import Tuple, List

from parsl import HighThroughputExecutor
from parsl.config import Config
from proxystore.connectors.file import FileConnector
from pytest import fixture, mark

from proxystore.store import Store
from proxystore.store import register_store
from proxystore.store import unregister_store

from colmena.queue.base import ColmenaQueues

from colmena.queue.python import PipeQueues
from colmena.models import ResourceRequirements
from .test_base import EchoTask, FakeMPITask
from colmena.task_server.parsl import ParslTaskServer


def f(x):
    return x + 1


def capitalize(y: List[str], x: str, **kwargs):
    return x.upper(), [i.lower() for i in y]


def bad_task(x):
    raise MemoryError()


def count_nodes(x, _resources: ResourceRequirements):
    return _resources.node_count


# Make the Parsl configuration. Use LocalThreads for Mac and Windows compatibility
@fixture()
def config(tmpdir):
    return Config(
        executors=[
            HighThroughputExecutor(max_workers_per_node=1, address='127.0.0.1')
        ],
        strategy=None,
        run_dir=str(tmpdir / 'run'),
    )


# Make a proxy store for larger objects
@fixture()
def store(tmpdir):
    connector = FileConnector(store_dir=str(tmpdir / 'proxy-store'))
    with Store('store', connector=connector, metrics=True) as store:
        register_store(store)
        yield store
        unregister_store(store.name)


@fixture(autouse=True)
def server_and_queue(config, store) -> Tuple[ParslTaskServer, ColmenaQueues]:
    queues = PipeQueues(proxystore_name='store', proxystore_threshold=5000, serialization_method='pickle')
    server = ParslTaskServer([(f, {'executors': 'all'}), capitalize, bad_task, EchoTask(), FakeMPITask(), count_nodes], queues, config)
    yield server, queues
    if server.is_alive():
        queues.send_kill_signal()
        server.join(timeout=30)


@mark.timeout(30)
def test_kill(server_and_queue):
    """Make sure the server shutdown signal works. Should not hang"""
    server, queue = server_and_queue
    queue.send_kill_signal()

    # Make sure it kills the server
    queue.send_kill_signal()
    server.start()
    server.join()

    assert server.exitcode == 0


@mark.timeout(30)
def test_run_simple(server_and_queue):
    """Make sure the run and stop """
    server, queue = server_and_queue

    # Start the service
    server.start()

    # Send a result and then check results
    queue.send_inputs(1, method='f')
    result = queue.get_result()
    assert result.success
    assert result.value == 2
    assert result.time.running > 0
    assert result.time.deserialize_inputs > 0
    assert result.time.serialize_results > 0
    assert result.timestamp.compute_started is not None
    assert result.timestamp.result_sent is not None


@mark.timeout(30)
def test_timeout(server_and_queue):
    """Make sure the timeout works"""
    server, _ = server_and_queue
    server.timeout = 1

    # Make sure the server dies
    server.start()
    server.join()


@mark.timeout(30)
def test_error_handling(server_and_queue):
    server, queue = server_and_queue

    # Start the server
    server.start()

    # Send a result and then get the error message back
    queue.send_inputs(None, method='f', keep_inputs=True)
    result = queue.get_result()
    assert result.args == (None,)
    assert result.value is None
    assert not result.success
    assert result.failure_info is not None
    assert result.time.running is not None

    # Send a task that kills the worker
    queue.send_inputs(None, method='bad_task')
    result = queue.get_result()
    assert 'MemoryError' in result.failure_info.exception


@mark.timeout(30)
def test_bash(server_and_queue):
    server, queue = server_and_queue

    # Get ready to receive tasks
    server.start()

    # Send a standard task
    queue.send_inputs(1, method='echo', keep_inputs=True)
    result = queue.get_result()
    assert result.success, result.failure_info
    assert result.value == '1\n'
    assert result.keep_inputs
    assert result.time.additional['exec_execution'] > 0
    assert result.inputs == ((1,), {})

    # Send an MPI task
    queue.send_inputs(1, method='fakempitask', keep_inputs=True)
    result = queue.get_result()
    assert result.success, result.failure_info
    assert result.value == '-N 1 -n 1 --cc depth echo -n 1\n'  # We're actually testing that it makes the correct command string
    assert result.keep_inputs
    assert result.time.additional['exec_execution'] > 0
    assert result.inputs == ((1,), {})

    # Send an MPI task
    queue.send_inputs(1, method='fakempitask', keep_inputs=True, resources=ResourceRequirements(node_count=2, cpu_processes=4))
    result = queue.get_result()
    assert result.success, result.failure_info
    assert result.value == '-N 8 -n 4 --cc depth echo -n 1\n'
    assert result.time.additional['exec_execution'] > 0
    assert result.inputs == ((1,), {})


@mark.timeout(30)
def test_resources(server_and_queue):
    server, queue = server_and_queue

    # Get ready to receive tasks
    server.start()

    # Launch a test with the default values
    queue.send_inputs(1, method='count_nodes')
    result = queue.get_result()
    assert result.success
    assert result.value == 1  # Default node count is 1

    # Launch a test with a user-specified value
    queue.send_inputs(1, method='count_nodes', resources={'node_count': 2})
    result = queue.get_result()
    assert result.success
    assert result.value == 2


@mark.timeout(30)
def test_proxy(server_and_queue, store):
    """Test a task that uses proxies"""

    # Start the server
    server, queue = server_and_queue
    server.start()

    # Send a big task
    big_string = "a" * 10000
    little_string = "A" * 10000
    queue.send_inputs([little_string], big_string, method='capitalize')
    result = queue.get_result()
    assert result.success, result.failure_info.exception
    assert len(result.time.proxy) == 3  # There are two proxies to resolve, one is created

    # Proxy the results ahead of time
    little_proxy = store.proxy(little_string)

    queue.send_inputs([little_proxy], big_string, method='capitalize')
    result = queue.get_result()
    assert result.success, result.failure_info.exception
    assert len(result.time.proxy) == 3

    # Try it with a kwarg
    queue.send_inputs(['a'], big_string, input_kwargs={'little': little_proxy}, method='capitalize',
                      keep_inputs=False)  # TODO (wardlt): test does not work with keep-inputs=True
    result = queue.get_result()
    assert result.success, result.failure_info.exception
    assert len(result.time.proxy) == 3


@mark.timeout(10)
def test_bad_method_name(server_and_queue):
    """Make sure tasks with undefined methods are returned with a meaningful error"""

    # Start the server
    server, queue = server_and_queue
    server.start()

    # Make sure it sends back a result
    queue.send_inputs(1, method='not_a_real_method')
    result = queue.get_result()
    assert not result.success
    assert 'not_a_real_method' in str(result.failure_info.exception)



"""Task server based on Globus Compute

Globus Compute provides the ability to execute functions on remote "endpoints" that provide access to
computational resources (e.g., cloud providers, HPC).
Tasks and results are communicated to/from the endpoint through a cloud service secured using Globus Auth."""

import logging
from typing import Dict, Callable, Optional, Tuple
from concurrent.futures import Future

from globus_compute_sdk import Client, Executor

from colmena.task_server.base import convert_to_colmena_method, FutureBasedTaskServer
from colmena.queue.python import PipeQueues

from colmena.models import Result

logger = logging.getLogger(__name__)


class GlobusComputeTaskServer(FutureBasedTaskServer):
    """Task server that uses Globus Compute to execute tasks on remote systems

    Create a task server by providing a dictionary of functions
    mapped to the `endpoint <https://funcx.readthedocs.io/en/latest/endpoints.html>`_
    on which each should run. The task server will wrap the provided function
    in an interface that tracks execution information (e.g., runtime) and
    `registers <https://funcx.readthedocs.io/en/latest/sdk.html#registering-functions>`_
    the wrapped function with Globus Compute.
    You must also provide a Globus Compute :class:`~globus_compute_sdk.client.Client`
    that the task server will use to authenticate with the web service.

    The task server works using Globus Compute's :class:`~globus_compute_sdk.executor.Executor`
    to communicate to the web service over a web socket.
    The functions used by the executor are registered when you create the task server,
    and the Executor is launched when you start the task server.
    """

    def __init__(self,
                 methods: Dict[Callable, str],
                 funcx_client: Client,
                 queues: PipeQueues,
                 timeout: Optional[int] = None,
                 batch_size: int = 128):
        """
        Args:
            methods: Map of functions to the endpoint on which it will run
            funcx_client: Authenticated Globus Compute client
            queues: Queues used to communicate with thinker
            timeout: Timeout for requests from the task queue
            batch_size: Maximum number of task request to receive before submitting
        """
        # Store the client that has already been authenticated.
        self.fx_client = funcx_client
        self.fx_exec: Optional[Executor] = None

        # Create a function with the latest version of the wrapper function
        self.registered_funcs: Dict[str, Tuple[str, str]] = {}  # Function name -> (funcX id, endpoints)
        for func, endpoint in methods.items():
            # Register a wrapped version of the function
            task = convert_to_colmena_method(func)
            func_fxid = self.fx_client.register_function(task)

            # Store the information for the function
            self.registered_funcs[task.name] = (func_fxid, endpoint)

        self._batch_options = dict(
            batch_size=batch_size,
        )

        # Initialize the outputs
        super().__init__(queues, self.registered_funcs.keys(), timeout)

    def perform_callback(self, future: Future, result: Result, topic: str):
        # Check if the failure was due to a ManagerLost
        #  TODO (wardlt): Remove when we have retry support in Globus Compute
        exc = future.exception()
        if 'Task failure due to loss of manager' in str(exc):
            logger.info('Caught an task that failed due to a lost manager. Resubmitting')
            self.process_queue(topic, result)
        else:
            super().perform_callback(future, result, topic)

    def _submit(self, task: Result, topic: str) -> Future:
        # Lookup the appropriate function ID and endpoint
        func, endp_id = self.registered_funcs[task.method]
        task.mark_start_task_submission()

        # set the executor's endpoint before submitting the task
        self.fx_exec.endpoint_id = endp_id
        logger.info(f'Submitting function {func} to run on {endp_id}')

        # Submit it to funcX to be executed
        future = self.fx_exec.submit_to_registered_function(func, kwargs={'result': task})

        logger.info(f'Submitted {task.method} to run on {endp_id}')
        return future

    def _setup(self):
        # Create an executor to asynchronously transmit funcX tasks and receive results
        self.fx_exec = Executor(funcx_client=self.fx_client,
                                batch_size=self._batch_options['batch_size'])

    def _cleanup(self):
        self.fx_exec.shutdown(cancel_futures=True)



"""Base classes used by Colmena to describe functions being executed by a workflow engine"""
import os
import shlex
import logging
import platform
from io import StringIO
from pathlib import Path
from subprocess import run
from tempfile import TemporaryDirectory
from time import perf_counter
from inspect import signature
from typing import Any, Dict, List, Tuple, Optional, Callable, Generator, Union, Iterable

from colmena.models.results import ResourceRequirements, Result, FailureInformation, WorkerInformation
from colmena.proxy import resolve_proxies_async, store_proxy_stats
from colmena.queue import ColmenaQueues


logger = logging.getLogger(__name__)


class ColmenaMethod:
    """Base wrapper for a Python function run as part of a Colmena workflow

    The wrapper handles the parts of running a Colmena task that are beyond running the function,
    such as serialization, timing, interfaces to ProxyStore.
    """

    name: str
    """Name used to identify the function"""

    @property
    def __name__(self):
        return self.name

    def function(self, *args, **kwargs) -> Any:
        """Function provided by the Colmena user"""
        raise NotImplementedError()

    def __call__(self, result: Result) -> Result:
        """Invoke a Colmena task request

        Args:
            result: Request, which inclues the arguments and will hold the result
        Returns:
            The input result object, populated with the results
        """
        # Mark that compute has started on the worker
        result.mark_compute_started()

        # Unpack the inputs
        result.time.deserialize_inputs = result.deserialize()

        # Start resolving any proxies in the input asynchronously
        start_time = perf_counter()
        input_proxies = []
        for arg in result.args:
            input_proxies.extend(resolve_proxies_async(arg))
        for value in result.kwargs.values():
            input_proxies.extend(resolve_proxies_async(value))
        result.time.async_resolve_proxies = perf_counter() - start_time

        # Add the worker information into the tasks, if available
        worker_info = {}
        # TODO (wardlt): Move this information into a separate, parsl-specific wrapper
        for tag in ['PARSL_WORKER_RANK', 'PARSL_WORKER_POOL_ID']:
            if tag in os.environ:
                worker_info[tag] = os.environ[tag]
        worker_info['hostname'] = platform.node()
        result.worker_info = WorkerInformation(**worker_info)

        # Determine additional kwargs to provide to the function
        additional_kwargs = {}
        for k, v in [('_resources', result.resources), ('_result', result)]:
            if k in result.kwargs:
                logger.warning(f'`{k}` provided as a kwargs. Unexpected things are about to happen')
            if k in signature(self.function).parameters:
                additional_kwargs[k] = v

        # Execute the function
        start_time = perf_counter()
        success = True
        try:
            output = self.function(*result.args, **result.kwargs, **additional_kwargs)
        except BaseException as e:
            output = None
            success = False
            result.failure_info = FailureInformation.from_exception(e)
        finally:
            end_time = perf_counter()

        # Store the results
        result.set_result(output, end_time - start_time)
        if not success:
            result.success = False
        result.mark_compute_ended()

        # Re-pack the results. Will store the proxy statistics
        result.time.serialize_results, _ = result.serialize()

        # Get the statistics for the proxy resolution
        for proxy in input_proxies:
            store_proxy_stats(proxy, result.time.proxy)

        return result


class PythonMethod(ColmenaMethod):
    """A Python function to be executed on a single worker

    Args:
        function: Generator function to be executed
        name: Name of the function. Defaults to `function.__name__`
    """

    function: Callable

    def __init__(self, function: Callable, name: Optional[str] = None) -> None:
        self.name = name or function.__name__
        self.function = function


class PythonGeneratorMethod(ColmenaMethod):
    """Python function which runs on a single worker and generates results iteratively

    Generator functions support streaming each iteration of the generator
    to the Thinker when a `streaming_queue` is provided.

    Args:
        function: Generator function to be executed
        name: Name of the function. Defaults to `function.__name__`
        store_return_value: Whether to capture the `return value <https://docs.python.org/3/reference/simple_stmts.html#the-return-statement>`_
            of the generator and store it in the Result object.
    """

    def __init__(self,
                 function: Callable[..., Union[Generator, Iterable]],
                 name: Optional[str] = None,
                 store_return_value: bool = False,
                 streaming_queue: Optional[ColmenaQueues] = None) -> None:
        self._function = function
        self.name = name or function.__name__
        self.store_return_value = store_return_value
        self.streaming_queue = streaming_queue

    def stream_result(self, y: Any, result: Result, start_time: float):
        """Send an intermediate result using the task queue

        Args:
            y: Yielded data from the generator function
            result: Result package carrying task metadata
            start_time: Start time of the algorithm, used to report
        """

        # Store the intermediate result in a copy of the input object
        result = result.model_copy(deep=True)
        result.set_result(
            y, perf_counter() - start_time, intermediate=True,
        )
        result.time.serialize_results, _ = result.serialize()

        # Send it back to the queue
        self.streaming_queue.send_result(result)

    def function(self, *args, _result: Result, **kwargs) -> Any:
        """Run the Colmena task and collect intermediate results to provide as a list"""

        # TODO (wardlt): Make push to task queue asynchronous
        gen = self._function(*args, **kwargs)
        iter_results = []
        start_time = perf_counter()
        while True:
            try:
                y = next(gen)
                if self.streaming_queue is None:
                    iter_results.append(y)
                else:
                    self.stream_result(y, _result, start_time)
            except StopIteration as e:
                if self.streaming_queue is not None:
                    if self.store_return_value:
                        return e.value
                elif self.store_return_value:
                    return iter_results, e.value
                else:
                    return iter_results


class ExecutableMethod(ColmenaMethod):
    """Task that involves running an executable using a system call.

    Such tasks often include a "pre-processing" step in Python that prepares inputs for the executable
    and a "post-processing" step which stores the outputs (either produced from stdout or written to files)
    as Python objects.

    Separating the task into these two functions and a system call for launching the program
    simplifies development (shorter functions that ar easier to test), and allows some workflow
    engines to improve performance by running processing and execution tasks separately.

    Implement a new ExecutableTask by defining the executable, a preprocessing method (:meth:`preprocess`),
    and a postprocessing method (:meth:`postprocess`).

    Use the ExecutableTask by instantiating a copy of your new class and then passing it to the task server
    as you would with any other function.

    **MPI Executables**

    Launching an MPI executable requires two parts: a path to an executable and a preamble defining how to launch it.
    Defining an MPI application using the instructions described above and then set the :attr:`mpi` attribute to ``True``.
    This will tell the Colmena task server to look for a "preamble" for how to launch the application.

    You may need to supply an MPI command invocation recipe for your particular cluster, depending on your choice of task server.
    Supply a template as the ``mpi_command_string`` field, which will be converted
    by `Python's string format function <https://docs.python.org/3/library/string.html#format-string-syntax>`_
    to produce a version of the command with the specific resource requirements of your task
    by the :meth:`render_mpi_launch` method.
    The attributes of this class (e.g., ``node_count``, ``total_ranks``) will be used as arguments to `format`.
    For example, a template of ``aprun -N {total_ranks} -n {cpu_process}`` will produce ``aprun -N 6 -n 3`` if you
    specify ``node_count=2`` and ``cpu_processes=3``.

    Args:
        executable: List of executable arguments
        name: Name used for the task. Defaults to ``executable[0]``
        mpi: Whether to use MPI to launch the exectuable
        mpi_command_string: Template for MPI launcher. See :attr:`mpi_command_string`.
    """

    executable: List[str]
    """Command used to launch the executable"""

    mpi: bool = False
    """Whether this is an MPI executable"""

    mpi_command_string: Optional[str] = None
    """Template string defining how to launch this application using MPI.
    Should include placeholders named after the fields in ResourceRequirements marked using {}'s.
    Example: `mpirun -np {total_ranks}`"""

    def __init__(self, executable: List[str], name: Optional[str] = None,
                 mpi: bool = False, mpi_command_string: Optional[str] = None) -> None:
        super().__init__()
        self.name = name or executable[0]
        self.executable = executable
        self.mpi = mpi
        self.mpi_command_string = mpi_command_string

    def render_mpi_launch(self, resources: ResourceRequirements) -> str:
        """Create an MPI launch command given the configuration

        Returns:
            MPI launch configuration
        """
        return self.mpi_command_string.format(total_ranks=resources.total_ranks,
                                              **resources.model_dump(exclude={'mpi_command_string'}))

    def preprocess(self, run_dir: Path, args: Tuple[Any], kwargs: Dict[str, Any]) -> Tuple[List[str], Optional[str]]:
        """Perform preprocessing steps necessary to prepare for executable to be started.

        These may include writing files to the local directory, creating CLI arguments,
        or standard input to be passed to the executable

        Args:
            run_dir: Path to a directory in which to write files used by an executable
            args: Arguments to the task, control how the run is set up
            kwargs: Keyword arguments to the function
        Returns:
            - Options to be passed as command line arguments to the executable
            - Values to pass to the standard in of the executable
        """
        raise NotImplementedError()

    def execute(self, run_dir: Path, arguments: List[str], stdin: Optional[str],
                resources: Optional[ResourceRequirements] = None) -> float:
        """Run an executable

        Args:
            run_dir: Directory in which to execute the code
            arguments: Command line arguments
            stdin: Content to pass in via standard in
            resources: Amount of resources to use for the application
        Returns:
            Runtime (unit: s)
        """

        # Make the shell command to be launched
        shell_cmd = self.assemble_shell_cmd(arguments, resources)
        logger.debug(f'Launching shell command: {" ".join(shell_cmd)}')

        # Launch it, routing the stdout and stderr as appropriate
        start_time = perf_counter()
        with open(run_dir / 'colmena.stdout', 'w') as fo, open(run_dir / 'colmena.stderr', 'w') as fe:
            if stdin is not None:
                stdin = StringIO(stdin)
            run(shell_cmd, stdout=fo, stderr=fe, stdin=stdin, cwd=run_dir)
        return perf_counter() - start_time

    def assemble_shell_cmd(self, arguments: List[str], resources: ResourceRequirements) -> List[str]:
        """Assemble the shell command to be launched

        Args:
            arguments: Command line arguments
            resources: Resource requirements
        Returns:
            Components of the shell command
        """

        # If resources are provided and the task is an MPI, generate the MPI executor
        if self.mpi:
            assert resources is not None, "Resources must be specified for MPI tasks"
            preamble = shlex.split(self.render_mpi_launch(resources))
        else:
            preamble = []

        # Get the full shell command
        shell_cmd = preamble + self.executable + arguments
        return shell_cmd

    def postprocess(self, run_dir: Path) -> Any:
        """Extract results after execution completes

        Args:
            run_dir: Run directory for the executable. Stdout will be written to `run_dir/colmena.stdout`
                    and stderr to `run_dir/colmena.stderr`
        """
        raise NotImplementedError()

    def function(self, *args, _resources: Optional[ResourceRequirements] = None, **kwargs):
        """Execute the function

        Args:
            args: Positional arguments
            kwargs: Keyword arguments
            _resources: Resources available. Optional. Only used for MPI tasks.
        """
        # Launch everything inside a temporary directory
        with TemporaryDirectory() as run_dir:
            run_dir = Path(run_dir)

            # Prepare the run directory
            cli_args, stdin = self.preprocess(run_dir, args, kwargs)

            # Execute everything
            self.execute(run_dir, cli_args, stdin, resources=_resources)

            # Return the post-processed results
            return self.postprocess(run_dir)



from typing import Any, Dict, Tuple, List, Optional
from pathlib import Path

from colmena.models.methods import ExecutableMethod


class EchoTask(ExecutableMethod):
    def __init__(self):
        super().__init__(executable=['echo'])

    def preprocess(self, run_dir: Path, args: Tuple[Any], kwargs: Dict[str, Any]) -> Tuple[List[str], Optional[str]]:
        return list(map(str, args)), None

    def postprocess(self, run_dir: Path) -> Any:
        return (run_dir / 'colmena.stdout').read_text()


class FakeMPITask(ExecutableMethod):
    def __init__(self):
        super().__init__(executable=['echo', '-n'],
                         name='fakempitask',
                         mpi=True,
                         mpi_command_string='echo -N {total_ranks} -n {cpu_processes} --cc depth')

    def preprocess(self, run_dir: Path, args: Tuple[Any], kwargs: Dict[str, Any]) -> Tuple[List[str], Optional[str]]:
        return list(map(str, args)), None

    def postprocess(self, run_dir: Path) -> Any:
        return (run_dir / 'colmena.stdout').read_text()



"""Test for ProxyStore utilities."""
import json

import pytest

from proxystore.proxy import Proxy
from proxystore.store import register_store
from proxystore.store import unregister_store
from proxystore.store.base import Store
from proxystore.connectors.local import LocalConnector

from colmena.proxy import get_store
from colmena.proxy import proxy_json_encoder
from colmena.proxy import ProxyJSONSerializationWarning
from colmena.proxy import resolve_proxies_async


class ExampleStore(Store):
    pass


@pytest.fixture
def proxy() -> Proxy:
    with Store('proxy-fixture-store', LocalConnector()) as store:
        yield store.proxy('test-value')


def test_get_store_already_registered() -> None:
    store = Store('test-store', LocalConnector())
    register_store(store)
    assert get_store('test-store') is store
    unregister_store(store)


def test_get_store_missing() -> None:
    assert get_store('test-store') is None


def test_get_store_initialize_from_config() -> None:
    # Create a temp store just to get the config
    config = Store('test-store', LocalConnector()).config()
    store = get_store('test-store', config)
    # Verify get_store registered the store globally by getting it again
    assert get_store('test-store') is store
    unregister_store(store)


def test_proxy_json_encoder() -> None:
    p = Proxy(lambda: 'test-value')
    with pytest.warns(ProxyJSONSerializationWarning):
        result = json.dumps({'proxy': p}, default=proxy_json_encoder)
    reconstructed_data = json.loads(result)
    assert 'proxy' in reconstructed_data


def test_proxy_json_encoder_no_proxies() -> None:
    # proxy_json_encoder should raise TypeError on non-Proxy types which
    # were unable to be serialized by the default serializer
    with pytest.raises(TypeError):
        json.dumps({'a': lambda: 1}, default=proxy_json_encoder)


def test_proxy_json_encoder_resolved_proxy(proxy) -> None:
    # force the proxy to resolve
    assert isinstance(proxy, str)
    result = json.dumps({'proxy': proxy}, default=proxy_json_encoder)
    assert result == '{"proxy": "test-value"}'


def test_resolve_proxy_async_object_arg(proxy) -> None:
    assert resolve_proxies_async(proxy) == [proxy]
    assert resolve_proxies_async('not a proxy') == []


def test_resolve_proxy_async_sequence_arg(proxy) -> None:
    assert resolve_proxies_async([proxy, 'not a proxy']) == [proxy]
    assert resolve_proxies_async((proxy, 'not a proxy')) == [proxy]


def test_resolve_proxy_async_dict_arg(proxy) -> None:
    assert resolve_proxies_async(
        {'proxy': proxy, 'other': 'not a proxy'},
    ) == [proxy]



from typing import Tuple, Generator
from parsl.executors import HighThroughputExecutor
from parsl.providers import LocalProvider
from parsl.config import Config
from colmena.queue.base import ColmenaQueues
from colmena.queue.python import PipeQueues
from colmena.task_server.parsl import ParslTaskServer

from pytest import fixture, mark

# Make global state for the retry task
RETRY_COUNT = 0


def retry_task(success_idx: int) -> bool:
    """Task that will succeed (return True) every `success_idx` times."""
    global RETRY_COUNT

    # If we haven't reached the success index, raise an error.
    if RETRY_COUNT < success_idx:
        RETRY_COUNT += 1
        raise ValueError('Retry')

    # Reset the retry count
    RETRY_COUNT = 0
    return True


@fixture
def reset_retry_count():
    """Reset the retry count before each test."""
    global RETRY_COUNT
    RETRY_COUNT = 0


@fixture()
def config(tmpdir):
    """Make the Parsl configuration."""
    # We must use a HighThroughputExecutor since the ThreadPoolExecutor
    # does not serialize the results objects within the task server.
    # This is necessary to test the retry policy, since otherwise the
    # failure information stored in the result (needed to check for retries)
    # is in shared memory and the perform_callback methods input result
    # will store the previous failure information.
    return Config(
        executors=[
            HighThroughputExecutor(
                    address='localhost',
                    label='htex',
                    max_workers_per_node=1,
                    provider=LocalProvider(init_blocks=1, max_blocks=1),
                ),
        ],
        strategy=None,
        run_dir=str(tmpdir / 'run'),
    )


@fixture
def server_and_queue(config) -> Generator[Tuple[ParslTaskServer, ColmenaQueues], None, None]:
    queues = PipeQueues()
    server = ParslTaskServer([retry_task], queues, config)
    yield server, queues
    if server.is_alive():
        queues.send_kill_signal()
        server.join(timeout=30)


@mark.timeout(10)
def test_retry_policy_max_retries_zero(server_and_queue, reset_retry_count):
    """Test the retry policy with max_retries=0"""

    # Start the server
    server, queue = server_and_queue
    server.start()

    # The task will fail every other time (setting success_idx=1)
    success_idx = 1

    for i in range(4):
        queue.send_inputs(success_idx, method='retry_task', max_retries=0)
        result = queue.get_result()
        assert result.success == (i % 2 == 1)
        if i % 2 == 1:
            assert result.value
            assert result.failure_info is None
        else:
            assert not result.success
            assert 'Retry' in str(result.failure_info.exception)


@mark.timeout(10)
@mark.parametrize(('success_idx', 'max_retries'), [(0, 0), (1, 1), (4, 10)])
def test_retry_policy_max_retries(server_and_queue, reset_retry_count, success_idx: int, max_retries: int):
    """Test the retry policy.

    This test checks the following cases:
    - A task that always succeeds (success_idx=0, max_retries=0)
    - A task that succeeds after one retry (success_idx=1, max_retries=1)
    - A task that succeeds after four retries (and extra max_retries) (success_idx=4, max_retries=10)
    """

    # Start the server
    server, queue = server_and_queue
    server.start()

    # Send the task to the queue
    queue.send_inputs(success_idx, method='retry_task', max_retries=max_retries)
    result = queue.get_result()
    assert result is not None
    assert result.success
    assert result.value
    assert result.failure_info is None
    assert result.retries == success_idx
    assert result.max_retries == max_retries



"""Queues that use Redis"""
from typing import Collection, Optional, Union, Dict, Any, Tuple
from uuid import uuid4
import logging

import redis

from colmena.exceptions import TimeoutException, KillSignalException
from colmena.models import SerializationMethod

from colmena.queue.base import ColmenaQueues

logger = logging.getLogger(__name__)


def _error_if_unconnected(f):
    def wrapper(queue: 'RedisQueues', *args, **kwargs) -> Any:
        if not queue.is_connected:
            raise ConnectionError('Not connected. Did you call `.connect()`?')
        return f(queue, *args, **kwargs)

    return wrapper


class RedisQueues(ColmenaQueues):
    """A basic redis queue for communications used by the task server

    A queue is defined by its prefix and a "topic" designation.
    The full list of available topics is defined when creating the queue,
    and simplifies writing software that waits for only certain types
    of messages without needing to manage several "queue" objects.
    By default, the :meth:`get` methods for the queue
    listen on all topics and the :meth:`put` method pushes to the default topic.
    You can put messages into certain "topical" queue and wait for responses
    that are from a single topic.

    The queue only connects when the `connect` method is called to avoid
    issues with passing an object across processes."""

    def __init__(self,
                 topics: Collection[str],
                 hostname: str = '127.0.0.1',
                 port: int = 6379,
                 prefix: str = uuid4(),
                 serialization_method: Union[str, SerializationMethod] = SerializationMethod.PICKLE,
                 keep_inputs: bool = True,
                 proxystore_name: Optional[Union[str, Dict[str, str]]] = None,
                 proxystore_threshold: Optional[Union[int, Dict[str, int]]] = None):
        """
        Args:
            hostname (str): Hostname of the Redis server
            port (int): Port on which to access Redis
            prefix (str): Name of the Redis queue
            topics: Names of topics that are known for this queue
            serialization_method: Method used to serialize task inputs and results
            keep_inputs: Whether to return task inputs with the result object
            proxystore_name (str, dict): Name of ProxyStore backend to use for all
                topics or a mapping of topic to ProxyStore backend for specifying
                backends for certain tasks. If a mapping is provided but a topic is
                not in the mapping, ProxyStore will not be used.
            proxystore_threshold (int, dict): Threshold in bytes for using
                ProxyStore to transfer objects. Optionally can pass a dict
                mapping topics to threshold to use different threshold values
                for different topics. None values in the mapping will exclude
                ProxyStore use with that topic.

        """
        super().__init__(topics, serialization_method, keep_inputs, proxystore_name, proxystore_threshold)
        self.hostname = hostname
        self.port = port
        self.redis_client = None
        self.prefix = prefix
        assert not any("{" in t for t in self.topics)
        self.connect()

    def __setstate__(self, state):
        super().__setstate__(state)

        # If you find the RedisClient placeholder, attempt to reconnect
        if self.redis_client == 'connected':
            self.redis_client = None
            self.connect()

    def __getstate__(self):
        state = super().__getstate__()

        # If connected, remove the unpicklable RedisClient and put a placeholder instead
        if self.is_connected:
            state['redis_client'] = 'connected'
        return state

    def connect(self):
        """Connect to the Redis server"""
        try:
            if not self.redis_client:
                self.redis_client = redis.StrictRedis(
                    host=self.hostname, port=self.port, decode_responses=True)
                self.redis_client.ping()  # Ping is needed to detect if connection failed
        except redis.exceptions.ConnectionError:
            logger.warning(f"ConnectionError while trying to connect to Redis@{self.hostname}:{self.port}")
            raise

    def disconnect(self):
        """Disconnect from the server

        Useful if sending the connection object to another process"""
        self.redis_client = None

    def _send_message(self, message, queue):
        """Sending a message to a certain redis queue

        Args:
            message: Message to send
            queue: Name of the queue
        """
        try:
            assert len(message) < 512 * 1024 * 1024, "Value too large for Redis. Task will fail"
            output = self.redis_client.rpush(queue, message)
            if not isinstance(output, int):
                raise ValueError(f'Message failed to push to {queue}')
        except redis.exceptions.ConnectionError:
            logger.warning(f"ConnectionError while trying to connect to Redis@{self.hostname}:{self.port}. Is Redis running?")
            raise

    def _get_message(self, queue: str, timeout: Optional[float]) -> str:
        """Get a message from a queue

        Args:
            queue: Name of the queue
            timeout: Timeout
        """

        # Redis uses 0 for non-blocking values
        if timeout is None:
            timeout = 0
        elif timeout < 1:
            logger.warning('Redis does not support timeouts less than 1s')
            timeout = 1

        try:
            output = self.redis_client.blpop(queue, timeout=int(timeout))
            if output is None:
                raise TimeoutException()
            return output[1]
        except redis.exceptions.ConnectionError:
            logger.warning(f"ConnectionError while trying to connect to Redis@{self.hostname}:{self.port}")
            raise

    @_error_if_unconnected
    def _send_request(self, message: str, topic: str):
        queue = f'{self.prefix}_requests'
        # Send it to the task server
        self._send_message(topic + message, queue)

    @_error_if_unconnected
    def _get_request(self, timeout: float = None) -> Tuple[str, str]:
        queue = f'{self.prefix}_requests'

        # Get the message
        message = self._get_message(queue, timeout)

        # The task data is JSON-encoded, so the first time a { appears is the beginning of the result
        #  the string before it is the topic
        if message.endswith("null"):
            raise KillSignalException()
        request_start = message.index("{")
        topic, request = message[:request_start], message[request_start:]

        return topic, request

    @_error_if_unconnected
    def _send_result(self, message: str, topic: str):
        queue = f'{self.prefix}_{topic}_result'
        self._send_message(message, queue)

    @_error_if_unconnected
    def _get_result(self, topic: str, timeout: int = None) -> str:
        queue = f'{self.prefix}_{topic}_result'
        return self._get_message(queue, timeout)

    @_error_if_unconnected
    def flush(self):
        """Flush the Redis queue"""

        all_queues = [f'{self.prefix}_requests']
        for topic in self.topics:
            all_queues.append(f'{self.prefix}_{topic}_result')
        try:
            self.redis_client.delete(*all_queues)
        except AttributeError:
            raise Exception("Queue is empty/flushed")
        except redis.exceptions.ConnectionError:
            logger.warning(f"ConnectionError while trying to connect to Redis@{self.hostname}:{self.port}")
            raise

    @property
    def is_connected(self):
        return self.redis_client is not None



"""Base classes for queues and related functions"""
import warnings
from abc import abstractmethod
from enum import Enum
from threading import Lock, Event
from typing import Optional, Tuple, Any, Collection, Union, Dict, Set
import logging

import proxystore.store

from colmena.models import Result, SerializationMethod, ResourceRequirements

logger = logging.getLogger(__name__)


class QueueRole(str, Enum):
    """Role a queue is used for"""

    ANY = 'any'
    SERVER = 'server'
    CLIENT = 'client'


class ColmenaQueues:
    """Base class for a queue used in Colmena.

    Follows the basic ``get`` and ``put`` semantics of most queues,
    with the addition of a "topic" used by Colmena to separate
    task requests or objects used for different purposes."""

    def __init__(self,
                 topics: Collection[str],
                 serialization_method: Union[str, SerializationMethod] = SerializationMethod.JSON,
                 keep_inputs: bool = True,
                 proxystore_name: Optional[Union[str, Dict[str, str]]] = None,
                 proxystore_threshold: Optional[Union[int, Dict[str, int]]] = None):
        """
        Args:
            topics: Names of topics that are known for this queue
            serialization_method: Method used to serialize task inputs and results
            keep_inputs: Whether to return task inputs with the result object
            proxystore_name (str, dict): Name of a registered ProxyStore
                `Store` instance. This can be a single name such that the
                corresponding `Store` is used for all topics or a mapping of
                topics to registered `Store` names. If a mapping is provided
                but a topic is not in the mapping, ProxyStore will not be used.
            proxystore_threshold (int, dict): Threshold in bytes for using
                ProxyStore to transfer objects. Optionally can pass a dict
                mapping topics to threshold to use different threshold values
                for different topics. None values in the mapping will exclude
                ProxyStore use with that topic.
        """

        # Store the list of topics and other simple options
        self.topics = set(topics)
        self.topics.add('default')
        self.keep_inputs = keep_inputs
        self.serialization_method = serialization_method
        self.role = QueueRole.ANY

        # Create {topic: proxystore_name} mapping
        self.proxystore_name = {t: None for t in self.topics}
        if isinstance(proxystore_name, str):
            self.proxystore_name = {t: proxystore_name for t in self.topics}
        elif isinstance(proxystore_name, dict):
            self.proxystore_name.update(proxystore_name)
        elif proxystore_name is not None:
            raise ValueError(f'Unexpected type {type(proxystore_name)} for proxystore_name')

        # Create {topic: proxystore_threshold} mapping
        self.proxystore_threshold = {t: None for t in self.topics}
        if isinstance(proxystore_threshold, int):
            self.proxystore_threshold = {t: proxystore_threshold for t in self.topics}
        elif isinstance(proxystore_threshold, dict):
            self.proxystore_threshold.update(proxystore_threshold)
        elif proxystore_threshold is not None:
            raise ValueError(f'Unexpected type {type(proxystore_threshold)} for proxystore_threshold')

        # Verify that ProxyStore backends exist
        for ps_name in set(self.proxystore_name.values()):
            if ps_name is None:
                continue
            store = proxystore.store.get_store(ps_name)
            if store is None:
                raise ValueError(
                    f'A Store with name "{ps_name}" has not been registered. '
                    'This is likely because the store needs to be '
                    'initialized prior to initializing the Colmena queue.'
                )

        # Log the ProxyStore configuration
        for topic in self.topics:
            ps_name = self.proxystore_name[topic]
            ps_threshold = self.proxystore_threshold[topic]

            if ps_name is None or ps_threshold is None:
                logger.debug(f'Topic {topic} will not use ProxyStore')
            else:
                logger.debug(
                    f'Topic {topic} will use ProxyStore backend "{ps_name}" '
                    f'with a threshold of {ps_threshold} bytes'
                )

        # Create a collection that holds the task which have been sent out, and an event that is triggered
        #  when the last task being sent out hits zero
        self._active_lock = Lock()
        self._active_tasks: Set[str] = set()
        self._all_complete = Event()

    def __getstate__(self):
        state = self.__dict__.copy()
        # We do not send the lock or event over pickle
        state.pop('_active_lock')
        state.pop('_all_complete')
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._active_lock = Lock()
        self._all_complete = Event()

    def _check_role(self, allowed_role: QueueRole, calling_function: str):
        """Check whether the queue is in an appropriate role for a requested function

        Emits a warning if the queue is in the wrong role

        Args:
            allowed_role: Role to check for
            calling_function: Name of the calling function
        """
        if self.role != QueueRole.ANY and self.role != allowed_role:
            warnings.warn(f'{calling_function} is intended for {allowed_role} not a {self.role}')

    @property
    def active_count(self) -> int:
        """Number of active tasks"""
        return len(self._active_tasks)

    def set_role(self, role: Union[QueueRole, str]):
        """Define the role of this queue.

        Controls whether users will be warned away from performing actions that are disallowed by
        a certain queue role, such as sending results from a client or issuing requests from a server"""
        role = QueueRole(role)
        self.role = role

    def get_result(self, topic: str = 'default', timeout: Optional[float] = None) -> Optional[Result]:
        """Get a completed result

        Args:
            topic: Which topic of task to wait for
            timeout: Timeout for waiting for a value
        Returns:
            (Result) Result from a computation
        Raises:
            TimeoutException if the timeout is met
        """
        self._check_role(QueueRole.CLIENT, 'get_result')

        # Get a value
        message = self._get_result(timeout=timeout, topic=topic)
        logger.debug(f'Received value: {str(message)[:25]}')

        # Parse the value and mark it as complete
        result_obj = Result.model_validate_json(message)
        result_obj.time.deserialize_results = result_obj.deserialize()
        result_obj.mark_result_received()

        # Some logging
        logger.info(f'Client received a {result_obj.method} result with topic {topic}')

        # Update the list of active tasks
        with self._active_lock:
            self._active_tasks.discard(result_obj.task_id)
            if len(self._active_tasks) == 0:
                self._all_complete.set()

        return result_obj

    def send_inputs(self,
                    *input_args: Any,
                    method: str = None,
                    input_kwargs: Optional[Dict[str, Any]] = None,
                    keep_inputs: Optional[bool] = None,
                    resources: Optional[Union[ResourceRequirements, dict]] = None,
                    topic: str = 'default',
                    task_info: Optional[Dict[str, Any]] = None,
                    max_retries: int = 0) -> str:
        """Send a task request

        Args:
            *input_args (Any): Positional arguments to a function
            method (str): Name of the method to run. Optional
            input_kwargs (dict): Any keyword arguments for the function being run
            keep_inputs (bool): Whether to override the
            topic (str): Topic for the queue, which sets the topic for the result
            resources: Suggestions for how many resources to use for the task
            task_info (dict): Any information used for task tracking
            max_retries (int): Maximum number of times to retry the task if it fails
        Returns:
            Task ID
        """
        self._check_role(QueueRole.CLIENT, 'send_inputs')

        # Make sure the queue topic exists
        if topic not in self.topics:
            raise ValueError(f'Unknown topic: {topic}. Known are: {", ".join(self.topics)}')

        # Make fake kwargs, if needed
        if input_kwargs is None:
            input_kwargs = dict()

        # Determine whether to override the default "keep_inputs"
        _keep_inputs = self.keep_inputs
        if keep_inputs is not None:
            _keep_inputs = keep_inputs

        # Gather ProxyStore info if we are using it with this topic
        ps_name = self.proxystore_name[topic]
        ps_threshold = self.proxystore_threshold[topic]
        ps_kwargs = {}
        if ps_name is not None and ps_threshold is not None:
            store = proxystore.store.get_store(ps_name)
            # proxystore_kwargs contains all the information we would need to
            # reconnect to the ProxyStore backend on any worker
            ps_kwargs.update({
                'proxystore_name': ps_name,
                'proxystore_threshold': ps_threshold,
                'proxystore_config': store.config(),
            })

        # Create a new Result object
        result = Result(
            (input_args, input_kwargs),
            method=method,
            keep_inputs=_keep_inputs,
            serialization_method=self.serialization_method,
            task_info=task_info,
            resources=resources or ResourceRequirements(),  # Takes either the user specified or a default,
            topic=topic,
            max_retries=max_retries,
            **ps_kwargs
        )

        # Push the serialized value to the task server
        result.time.serialize_inputs, proxies = result.serialize()
        self._send_request(result.json(exclude_none=True), topic)
        logger.info(f'Client sent a {method} task with topic {topic}. Created {len(proxies)} proxies for input values')

        # Store the task ID in the active list
        with self._active_lock:
            self._active_tasks.add(result.task_id)
            self._all_complete.clear()
        return result.task_id

    def wait_until_done(self, timeout: Optional[float] = None):
        """Wait until all out-going tasks have completed

        Returns:
            Whether the event was set within the timeout
        """
        self._check_role(QueueRole.CLIENT, 'wait_until_done')
        return self._all_complete.wait(timeout=timeout)

    def get_task(self, timeout: float = None) -> Tuple[str, Result]:
        """Get a task object

        Args:
            timeout (float): Timeout for waiting for a task
        Returns:
            - (str) Topic of the calculation. Used in defining which queue to use to send the results
            - (Result) Task description
        Raises:
            TimeoutException: If the timeout on the queue is reached
            KillSignalException: If the queue receives a kill signal
        """
        self._check_role(QueueRole.SERVER, 'get_task')

        # Pull a record off of the queue
        topic, message = self._get_request(timeout)
        logger.debug(f'Received a task message with topic {topic} inbound queue')

        # Get the message
        task = Result.model_validate_json(message)
        task.mark_input_received()
        return topic, task

    def send_kill_signal(self):
        """Send the kill signal to the task server"""
        self._check_role(QueueRole.CLIENT, 'send_kill_signal')
        self._send_request("null", topic='default')

    def send_result(self, result: Result):
        """Send a value to a client

        Args:
            result (Result): Result object to communicate back
            topic (str): Topic of the calculation
        """
        self._check_role(QueueRole.SERVER, 'send_result')
        result.mark_result_sent()
        self._send_result(result.json(), topic=result.topic)

    @abstractmethod
    def _get_request(self, timeout: int = None) -> Tuple[str, str]:
        """Get a task request from the client

        Args:
            timeout: Timeout for the blocking get in seconds
        Returns:
            - (str) Topic of the item
            - (str) Serialized version of the task request object
        Raises:
            (TimeoutException) if the timeout is reached
            (KillSignalException) if a kill signal (the string ``null``) is received
        """
        pass

    @abstractmethod
    def _send_request(self, message: str, topic: str):
        """Push a task request to the task server

        Args:
            message (str): JSON-serialized version of the task request object
            topic (str): Topic of the queue
        """
        pass

    @abstractmethod
    def _get_result(self, topic: str, timeout: int = None) -> str:
        """Get a result from the task server

        Returns:
            A serialized form of the result method
        Raises:
            (TimeoutException) if the timeout is reached
        """
        pass

    @abstractmethod
    def _send_result(self, message: str, topic: str):
        """Push a result object from task server to thinker

        Args:
            message (str): Serialized version of the task request object
            topic (str): Topic of the queue
        """

    @abstractmethod
    def flush(self):
        """Remove all existing results from the queues"""
        pass



"""Test the Thinker class"""
import logging
from threading import Event
from time import sleep

from pytest import fixture, mark

from colmena.models import Result
from colmena.queue.python import PipeQueues
from colmena.thinker import BaseThinker, agent, result_processor, task_submitter, event_responder
from colmena.thinker.resources import ResourceCounter


class ExampleThinker(BaseThinker):

    def __init__(self, queues, rec, flag: Event, daemon: bool):
        super().__init__(queues, rec, daemon=daemon)
        self.flag = flag
        self.func_ran = False
        self.last_value = None
        self.submitted = None
        self.event = Event()
        self.event_responded = False
        self.n_slots = 1
        self.teardown_result = None

    def prepare_agent(self):
        if self.agent_name == "setup_teardown":
            self.local_details.payload = 'hello'

    def tear_down_agent(self):
        if self.agent_name == "setup_teardown":
            self.teardown_result = self.local_details.payload + "_goodbye"

    @agent(startup=True)
    def startup_function(self):
        self.func_ran = True

    @agent(startup=True)
    def setup_teardown(self):
        """Agent that tests the setup and teardown logic"""
        pass

    @agent
    def function(self):
        self.flag.wait()

    @result_processor
    def process_results(self, result: Result):
        self.last_value = result.value

    @task_submitter(n_slots='n_slots')  # Look it up from the class attribute at runtime
    def submit_task(self):
        assert self.rec.available_slots(None) >= 0
        self.submitted = True
        sleep(0.1)  # Used to prevent a race condition with responder. By the time this event finishes

    @event_responder(event_name='event', reallocate_resources=True, max_slots='n_slots',
                     gather_from=None, gather_to="event", disperse_to="event")
    def responder(self):
        self.rec.acquire("event", 1)
        self.event_responded = True
        self.rec.release("event", 1)


@fixture()
def queues():
    return PipeQueues([])


def test_detection():
    # Test that the wrappers set the appropriate attributes
    assert hasattr(ExampleThinker.startup_function, '_colmena_agent')
    assert getattr(ExampleThinker.startup_function, '_colmena_startup')
    assert hasattr(ExampleThinker.function, '_colmena_agent')
    assert not getattr(ExampleThinker.function, '_colmena_startup')
    assert hasattr(ExampleThinker.process_results, '_colmena_agent')
    assert hasattr(ExampleThinker.submit_task, '_colmena_agent')
    assert hasattr(ExampleThinker.responder, '_colmena_agent')

    # Test detecting the agents
    assert len(ExampleThinker.list_agents()) == 6
    assert 'function' in [a.__name__ for a in ExampleThinker.list_agents()]


def test_logger_name(queues):
    """Test the name of loggers"""

    class SimpleThinker(BaseThinker):
        pass

    # See if the default name holds
    thinker = SimpleThinker(queues)
    assert 'SimpleThinker' in thinker.logger.name

    # See if we can provide it its own name
    thinker = SimpleThinker(queues, logger_name='my_logger')
    assert thinker.logger.name == 'my_logger'


@mark.timeout(5)
def test_logger_timings_event(queues, caplog):
    class TestThinker(BaseThinker):
        event: Event = Event()

        @event_responder(event_name='event')
        def a(self):
            self.done.set()

        @event_responder(event_name='event')
        def b(self):
            self.done.set()

    # Start the thinker
    with caplog.at_level(logging.INFO):
        thinker = TestThinker(queues, daemon=True)
        thinker.start()
        sleep(0.5)
        thinker.event.set()
        thinker.join(timeout=1)
    assert thinker.done.is_set()
    assert any('Runtime' in record.msg for record in caplog.records if '.a' in record.name)
    assert sum('All responses to event complete' in record.msg for record in caplog.records) == 1


@mark.timeout(5)
def test_logger_timings_process(queues, caplog):
    class TestThinker(BaseThinker):

        @result_processor()
        def process(self, _):
            self.done.set()

    # Start the thinker
    thinker = TestThinker(queues, daemon=True)
    thinker.start()

    # Thinker sets queue roles to "client". Switch back to "any" role
    # since we use methods intended for the "server" role in the test.
    queues.set_role("any")

    # Spoof a result completing
    queues.send_inputs(1, method='test')
    topic, result = queues.get_task()
    result.set_result(1, 1)
    with caplog.at_level(logging.INFO):
        queues.send_result(result)

        # Wait then check the logs
        sleep(0.5)

    assert thinker.done.is_set()
    assert any('Runtime' in record.msg for record in caplog.records if '.process' in record.name), caplog.record_tuples


@mark.timeout(5)
def test_logger_timings_submitter(queues, caplog):
    class TestThinker(BaseThinker):

        @task_submitter()
        def submit(self):
            self.done.set()

    # Start the thinker
    with caplog.at_level(logging.INFO):
        thinker = TestThinker(queues, ResourceCounter(1), daemon=True)
        thinker.start()
        sleep(0.5)

    assert thinker.done.is_set()
    assert any('Runtime' in record.msg for record in caplog.records if '.submit' in record.name), caplog.record_tuples


@mark.timeout(5)
def test_run(queues):
    """Test the behavior of all agents"""
    # Make the server and thinker
    flag = Event()
    rec = ResourceCounter(1, ["event"])
    rec.acquire(None, 1)
    th = ExampleThinker(queues, rec, flag, daemon=True)

    # Thinker sets queue roles to "client". Switch back to "any" role
    # since we use methods intended for the "server" role in the test.
    queues.set_role("any")

    # Launch it and wait for it to run
    th.start()

    # Make sure startup function ran, and the thinker did not stop
    sleep(.1)
    assert th.func_ran
    assert not flag.is_set()
    assert th.is_alive()
    assert not th.done.is_set()

    # Make sure the setup and teardown logic worked
    assert th.teardown_result == "hello_goodbye"

    # Test task processor: Push a result to the queue and make sure it was received
    queues.send_inputs(1)
    topic, task = queues.get_task()
    task.set_result(4)
    queues.send_result(task)
    sleep(0.1)
    assert th.last_value == 4

    # Test task submitter: Release the nodes and see if it submits
    assert not th.submitted
    th.rec.release(None, 1)
    sleep(0.1)
    assert th.is_alive()
    assert th.submitted
    assert rec.available_slots(None) == 0

    # Test event responder: Trigger event, see if it triggers and acquires resources
    assert len(th.barriers) == 1
    assert th.barriers['event'].n_waiting == 0
    assert not th.event.is_set()
    assert not th.event_responded
    th.event.set()
    sleep(0.1)  # Give enough time for this thread to start up before releasing resources
    for _ in range(3):
        th.rec.release(None, 1)  # Repeat as submit_task/responder are competing for resources
    sleep(0.1)
    assert th.is_alive()
    assert th.event_responded
    assert th.rec.available_slots("event") >= 1
    assert not th.event.is_set()

    # Set the "finish" flag after sending a result out
    queues.send_inputs(1)
    flag.set()
    sleep(1)
    assert th.is_alive()

    # The system should not exit until all results are back
    topic, task = queues.get_task()
    task.set_result(4)
    queues.send_result(task)
    assert th.queues.wait_until_done(timeout=2)
    sleep(0.1)
    assert not th.is_alive()


@mark.timeout(5)
def test_exception(queues):
    """Verify that thinkers stop properly with an exception"""

    # Make a thinker that will fail on startup
    class BadThinker(BaseThinker):

        def __init__(self, queues):
            super().__init__(queues)
            self.flag = Event()
            self.was_set = False

        @event_responder(event_name='flag')
        def do_nothing(self):
            self.was_set = False
            pass

        @agent(startup=True)
        def fail(self):
            raise ValueError()

    # Will only exit within a timeout if the exception is properly set
    thinker = BadThinker(queues)
    thinker.run()



"""Test the base class using the thread queue implemetnation"""
from pytest import raises, fixture, warns

from colmena.models import SerializationMethod
from colmena.queue.base import ColmenaQueues
from colmena.queue.python import PipeQueues


class Test:
    """Class used in a serialization test"""
    x = None


@fixture()
def queue() -> ColmenaQueues:
    return PipeQueues(['a', 'b'])


def test_role(queue):
    """Test defining queue roles"""
    assert queue.role == 'any'

    queue.set_role('client')
    assert queue.role == 'client'

    with raises(ValueError):
        queue.set_role('asdf')

    # No warning here
    queue.send_inputs('test')

    # Make sure a warning is created
    with warns() as w:
        queue.get_task()
    assert 'get_task' in w[0].message.args[0]


def test_requests(queue):
    """Test sending requests"""

    # Send a request with all the fixin's
    queue.send_inputs('ab', method='test_method', input_kwargs={'1': 1}, task_info={'info': 'important'}, topic='a')

    # Make sure it comes through
    topic, request = queue.get_task()
    request.deserialize()
    assert request.serialization_method == queue.serialization_method
    assert request.keep_inputs == queue.keep_inputs
    assert topic == 'a'
    assert request.args == ('ab',)
    assert request.kwargs == {'1': 1}
    assert request.method == 'test_method'
    assert request.task_info == {'info': 'important'}

    # Make sure we can change the defaults
    queue.send_inputs('cd', method='test_method', keep_inputs=not queue.keep_inputs)
    topic, request = queue.get_task()
    request.deserialize()
    assert request.keep_inputs != queue.keep_inputs


def test_results(queue):
    """Test receiving results"""

    # Send a request and get it
    queue.send_inputs('ab')
    topic, request = queue.get_task()
    request.deserialize()
    request.set_result(1)
    assert topic == 'default'

    # Set a result value and send it back
    request.serialize()
    queue.send_result(request)
    result = queue.get_result(topic=topic)
    assert result.value == 1


def test_serialization(queue):
    """Test communicating results that must be pickled"""
    # Attempt to push a non-JSONable object to the queue
    queue.send_inputs(Test(), method='test_method', keep_inputs=True)
    topic, task = queue.get_task()
    assert isinstance(task.inputs[0][0], str)
    task.deserialize()
    assert task.args[0].x is None

    # Set the value
    # Test sending the value back
    x = Test()
    x.x = 1
    task.set_result(x)
    task.serialize()
    queue.send_result(task)
    result = queue.get_result(topic=topic)
    assert result.args[0].x is None
    assert result.value.x == 1


def test_clear_inputs(queue):
    """Test clearing the inputs after storing the result"""

    # Sent a method request
    queue.keep_inputs = False
    queue.send_inputs(1)
    _, result = queue.get_task()
    result.deserialize()
    result.set_result(1)

    # Make sure the inputs were deleted
    assert result.args == ()

    # Make sure we can override it, if desired
    queue.send_inputs(1, keep_inputs=True)
    _, result = queue.get_task()
    result.deserialize()
    result.set_result(1)

    assert result.args == (1,)


def test_pickling_error(queue):
    """Test communicating results that need to be pickled fails without correct setting"""
    queue.serialization_method = SerializationMethod.JSON
    # Attempt to push a non-JSON-able object to the queue
    with raises(TypeError):
        queue.send_inputs(Test())


def test_task_info(queue):
    """Make sure task info gets passed along"""

    # Sent a method request
    queue.send_inputs(1, task_info={'id': 'test'})
    topic, result = queue.get_task()
    result.deserialize()
    result.set_result(1)

    # Send it back
    result.serialize()
    queue.send_result(result)
    result = queue.get_result()
    assert result.task_info == {'id': 'test'}


def test_resources(queue):
    # Test with defaults
    queue.send_inputs(1, method='test')
    topic, result = queue.get_task()
    assert result.resources.node_count == 1

    # Test with non-defaults
    queue.send_inputs(1, resources={'node_count': 2})
    topic, result = queue.get_task()
    assert result.resources.node_count == 2


def test_event_count(queue):
    # Sent a method request
    task_id = queue.send_inputs(1)
    assert queue.active_count == 1
    assert not queue.wait_until_done(timeout=1)

    # Make sure the task ID does not change
    topic, task = queue.get_task()
    assert task_id == task.task_id

    # Sent the task back
    task.set_result(1)
    queue.send_result(task)
    queue.get_result()
    assert queue.active_count == 0
    assert queue.wait_until_done(timeout=1)

    # Send another and make sure the event is reset
    queue.send_inputs(1)
    assert not queue.wait_until_done(timeout=1)



"""Queues built on Python's native libraries"""
from multiprocessing.connection import Connection
from typing import Collection, Union, Optional, Dict, Tuple
from multiprocessing import Pipe

from colmena.exceptions import TimeoutException, KillSignalException
from colmena.queue.base import ColmenaQueues
from colmena.models import SerializationMethod


class PipeQueues(ColmenaQueues):
    """Queues using Python's implementation
    of `multiprocessing Pipes <https://docs.python.org/3/library/multiprocessing.html#exchanging-objects-between-processes>`_"""

    def __init__(self,
                 topics: Collection[str] = (),
                 serialization_method: Union[str, SerializationMethod] = SerializationMethod.PICKLE,
                 keep_inputs: bool = True,
                 proxystore_name: Optional[Union[str, Dict[str, str]]] = None,
                 proxystore_threshold: Optional[Union[int, Dict[str, int]]] = None):
        super().__init__(topics, serialization_method, keep_inputs, proxystore_name, proxystore_threshold)

        # Make the queues
        self.request_queue = Pipe(duplex=False)
        self.result_queues = dict((t, Pipe(duplex=False)) for t in self.topics)

    def _send_request(self, message: str, topic: str):
        self.request_queue[1].send((topic, message))

    def _get_request(self, timeout: int = None) -> Optional[Tuple[str, str]]:
        if self.request_queue[0].poll(timeout):
            msg = self.request_queue[0].recv()
            if msg[1] == "null":
                raise KillSignalException()
            return msg
        else:
            raise TimeoutException()

    def _send_result(self, message: str, topic: str):
        assert len(message) < 32 * 1024 * 1024, "Messages larger than 32 MiB may fail"
        self.result_queues[topic][1].send(message)

    def _get_result(self, topic: str, timeout: int = None) -> str:
        if self.result_queues[topic][0].poll(timeout):
            return self.result_queues[topic][0].recv()
        else:
            raise TimeoutException()

    def flush(self):
        def _flush(pipe: Tuple[Connection, Connection]):
            """Pull from a queue until it is empty"""
            while pipe[0].poll():
                pipe[0].recv()

        _flush(self.request_queue)
        for v in self.result_queues.values():
            _flush(v)



from typing import Tuple, List, Optional, Dict, Any, Iterator
from functools import lru_cache, update_wrapper
from pathlib import Path
from math import isnan

from pytest import fixture
from proxystore.connectors.file import FileConnector
from proxystore.store import Store
from proxystore.store import register_store
from proxystore.store import unregister_store

from colmena.models.methods import ExecutableMethod, PythonMethod, PythonGeneratorMethod
from colmena.models import ResourceRequirements, Result, SerializationMethod
from colmena.queue import PipeQueues


class EchoTask(ExecutableMethod):
    def __init__(self):
        super().__init__(executable=['echo'])

    def preprocess(self, run_dir: Path, args: Tuple[Any], kwargs: Dict[str, Any]) -> Tuple[List[str], Optional[str]]:
        return list(map(str, args)), None

    def postprocess(self, run_dir: Path) -> Any:
        return (run_dir / 'colmena.stdout').read_text()


@fixture
def result() -> Result:
    """Result object ready to be executed by a Task"""
    result = Result.from_args_and_kwargs((1,), method='echo')
    result.serialize()
    return result


@fixture
def store(tmpdir):
    with Store('store', FileConnector(tmpdir), metrics=True) as store:
        register_store(store)
        yield store
        unregister_store(store)


def echo(x: Any) -> Any:
    return x


def generator(i: int) -> Iterator[int]:
    """Generator which iterates over a list then returns a string saying "done" """
    yield from range(i)
    return "done"


def test_python_task(result):
    """A standard Python function"""

    # Make sure name resolution functions as intended
    task = PythonMethod(function=echo)
    assert task.name == 'echo'

    task = PythonMethod(function=echo, name='test')
    assert task.name == 'test'

    # Ensure it sets the task timings while running
    assert isnan(result.timestamp.compute_started)
    result = task(result)
    assert not isnan(result.timestamp.compute_ended)

    assert isinstance(result.value, str)  # Data is still serialized
    result.deserialize()
    assert result.value == 1


def test_generator(result):
    task = PythonGeneratorMethod(function=generator, store_return_value=False)
    assert task.name == 'generator'

    result = task(result)
    assert result.success, result.failure_info.traceback
    result.deserialize()
    assert result.value == [0,]


def test_generator_with_return(result):
    task = PythonGeneratorMethod(function=generator, name='with_return', store_return_value=True)
    assert task.name == 'with_return'

    result = task(result)
    assert result.success, result.failure_info.traceback
    result.deserialize()
    assert result.value == [[0,], 'done']


def test_generator_streaming(result):
    """Trigger streaming by adding a queue to the task definition"""

    queue = PipeQueues()
    task = PythonGeneratorMethod(function=generator, name='stream', store_return_value=True, streaming_queue=queue)

    result.topic = 'default'
    result = task(result)
    assert result.success, result.failure_info.traceback
    result.deserialize()
    assert result.value == 'done'
    assert result.complete

    intermediate = queue.get_result(timeout=1)
    assert intermediate.success
    intermediate.deserialize()
    assert not intermediate.complete
    assert intermediate.value == 0

    assert result.time.running >= intermediate.time.running


def test_executable_task(result):
    # Run a basic task
    task = EchoTask()
    assert task.name == 'echo'
    assert task.executable == ['echo']

    result = task(result)
    result.deserialize()
    assert result.success, result.failure_info.traceback
    assert result.value == '1\n'

    # Run an "MPI task"
    task.mpi = True
    task.mpi_command_string = 'aprun -N {total_ranks} -n {cpu_processes} --cc depth'
    assert task.render_mpi_launch(ResourceRequirements(node_count=2, cpu_processes=4)) == 'aprun -N 8 -n 4 --cc depth'

    task.mpi_command_string = 'echo -N {total_ranks} -n {cpu_processes} --cc depth'
    result.resources = ResourceRequirements(node_count=2, cpu_processes=3)
    result.serialize()
    result = task(result)
    result.deserialize()
    assert result.value == '-N 6 -n 3 --cc depth echo 1\n'


def test_run_function(store):
    """Make sure the run function behaves as expected:

    - Records runtimes
    - Tracks proxy statistics
    """

    # Make the result and configure it to use the store
    result = Result(inputs=(('a' * 1024,), {}))
    result.proxystore_name = store.name
    result.proxystore_threshold = 128
    result.proxystore_config = store.config()

    # Serialize it
    result.serialization_method = SerializationMethod.PICKLE
    result.serialize()

    # Run the function
    task = PythonMethod(lambda x: x.upper(), name='upper')
    result = task(result)

    # Make sure the timings are all set
    assert result.time.running > 0
    assert result.time.async_resolve_proxies > 0
    assert result.time.deserialize_inputs > 0
    assert result.time.serialize_results > 0
    assert result.timestamp.compute_ended >= result.timestamp.compute_started

    # Make sure we have stats for both proxies
    assert len(result.time.proxy) == 2
    assert all('store.proxy' in v['times'] for v in result.time.proxy.values())


def test_with_decorator(result):
    """Ensure streaming still works with decorators"""

    dec = lru_cache()(generator)
    update_wrapper(dec, generator)
    task = PythonGeneratorMethod(function=generator, store_return_value=True)
    assert task.name == 'generator'

    result = task(result)
    assert result.success, result.failure_info.traceback
    result.deserialize()
    assert result.value == [[0], 'done']



"""Models used for different aspects of a Colmena application"""

from colmena.models.results import Result, ResourceRequirements, FailureInformation, SerializationMethod

__all__ = ['Result', 'ResourceRequirements', 'FailureInformation', 'ExecutableTask', 'SerializationMethod']



from threading import Timer, Event
from time import sleep

from pytest import fixture, mark

from colmena.thinker.resources import ResourceCounter, ReallocatorThread


@fixture
def rec() -> ResourceCounter:
    return ResourceCounter(8, ['ml', 'sim'])


def test_initialize(rec):
    assert rec.unallocated_slots == 8
    assert rec.available_slots(None) == 8


def test_allocations(rec):
    # Move 8 nodes to the "ml" pool
    assert rec.reallocate(None, "ml", 8)
    assert rec.unallocated_slots == 0
    assert rec.available_slots("ml") == 8

    # Checkout all of them
    assert rec.acquire("ml", 8, timeout=1)
    assert rec.available_slots("ml") == 0
    assert rec.allocated_slots("ml") == 8

    # Request unavailable nodes to test a timeout
    assert not rec.acquire("ml", 1, timeout=0.02)

    # Release nodes
    assert rec.release("ml", 4, rerequest=False) is None
    assert rec.available_slots("ml") == 4

    # Release and re-request
    assert rec.release("ml", 4, rerequest=True, timeout=1)
    assert rec.available_slots("ml") == 4
    assert rec.available_slots("ml") == 4

    # Attempt a transfer that times out
    assert not rec.reallocate("ml", "sim", n_slots=5, timeout=0.01)
    assert rec.available_slots("ml") == 4

    # Attempt a transfer that completes
    assert rec.reallocate("ml", "sim", n_slots=4, timeout=4)
    assert rec.available_slots("sim") == 4
    assert rec.available_slots("ml") == 0
    assert rec.allocated_slots("sim") == 4
    assert rec.unallocated_slots == 0

    # Test out the "cancel if" events
    stop = Event()
    stop.set()
    assert not rec.acquire("sim", n_slots=5, cancel_if=stop)
    assert rec.available_slots("sim") == 4

    stop.clear()
    t = Timer(0.2, function=stop.set)
    t.start()
    assert not rec.acquire("sim", n_slots=5, cancel_if=stop)
    assert rec.available_slots("sim") == 4
    assert stop.is_set()

    # Test out reallocate all
    rec.reallocate("sim", "ml", "all")
    assert rec.allocated_slots("sim") == 0
    assert rec.allocated_slots("ml") == 8

    # Test out non-blocking re-allocator
    trigger = Event()
    rec.reallocate("ml", "sim", 8, block=False, callback=trigger.set)  # Will not immediately complete as there are reserved ML slotes
    sleep(0.1)
    assert not trigger.is_set()
    assert rec.allocated_slots("ml") == 8

    rec.release("ml", 4)  # Release so that the reallocation can complete
    sleep(0.1)
    assert trigger.is_set()
    assert rec.allocated_slots("sim") == 8


def test_reallocator(rec):
    # Start with everything allocated to "simulation"
    rec.reallocate(None, "sim", 8)

    # Test allocating up to the maximum
    stop = Event()
    alloc = ReallocatorThread(rec, stop_event=stop, gather_from="sim", gather_to="ml", disperse_to=None, max_slots=2)
    alloc.start()
    sleep(0.2)
    assert alloc.is_alive()
    assert rec.allocated_slots("ml") == 2

    # Once you set "stop," the thread should move resources to "unallocated"
    stop.set()
    sleep(0.2)
    assert not alloc.is_alive()
    assert rec.unallocated_slots == 2

    # Test without a maximum allocation
    stop.clear()
    alloc = ReallocatorThread(rec, stop_event=stop, gather_from="sim", gather_to="ml", disperse_to=None)
    alloc.start()
    sleep(0.2)
    assert alloc.is_alive()
    assert rec.available_slots("sim") == 0

    # Once you trigger the "stop event," the thread should move all resources to "None" and then exit
    stop.set()
    sleep(2)  # We check for the flag every 1s
    assert not alloc.is_alive()
    assert rec.unallocated_slots == 8


@mark.timeout(2)
@mark.repeat(4)
def test_reallocator_deadlock(rec):
    """Creates the deadlock reported in https://github.com/exalearn/colmena/issues/43"""
    rec.reallocate(None, "sim", 8)
    assert rec.available_slots("sim") == 8

    # Create two allocators: One that pulls from sim and another that pulls from ml
    ml_alloc = ReallocatorThread(rec, gather_from="sim", gather_to="ml", disperse_to="sim", max_slots=8)
    sim_alloc = ReallocatorThread(rec, gather_from="ml", gather_to="sim", disperse_to="ml", max_slots=8)

    # Start the ML allocator, which will pull resources from sim to ml
    ml_alloc.start()
    sleep(0.010)  # Give enough time for allocation to finish
    assert not ml_alloc.stop_event.is_set()
    assert rec.available_slots("ml") == 8
    assert rec.acquire("ml", 8)
    assert rec.available_slots("ml") == 0

    # Start the sim allocator, which will ask to pull resources from ml over to sim
    sim_alloc.start()
    sleep(0.001)
    assert rec.available_slots("ml") == 0

    # Send the stop signal and wait for ml_alloc to exit
    ml_alloc.stop_event.set()
    rec.release("ml", 8)
    ml_alloc.join(1)
    assert not ml_alloc.is_alive()



"""Tests for the data models"""
import sys

from colmena.models import ResourceRequirements, Result


def test_resources():
    res = ResourceRequirements(node_count=2, cpu_processes=4)
    assert res.total_ranks == 8


def test_message_sizes():
    result = Result(
        (('0' * 8,), {'test': '0' * 8}),
        method='test',
        keep_inputs=False,
        serialization_method='json',
    )

    # Make sure the size of the inputs is stored
    result.serialize()
    assert result.message_sizes['inputs'] >= 2 * sys.getsizeof('0' * 8)
    assert 'value' not in result.message_sizes

    # Add a result
    result.deserialize()
    result.set_result(1, 1)
    result.serialize()
    assert result.message_sizes['inputs'] >= 2 * sys.getsizeof('0' * 8)
    assert result.message_sizes['inputs'] >= sys.getsizeof(1)



"""Test local task server"""

from pytest import fixture

from colmena.queue.python import PipeQueues
from colmena.task_server.local import LocalTaskServer


def f(x):
    return x


@fixture
def queues():
    return PipeQueues()


@fixture
def server(queues):
    server = LocalTaskServer(queues, methods=[f], num_workers=2)
    server.start()
    yield server
    queues.send_kill_signal()
    server.join()


def test_local_service(server, queues):
    queues.send_inputs(1, method='f')

    result = queues.get_result(timeout=4)
    assert result.success, result.failure_info.traceback
    assert result.value == 1



"""Utilities for interacting with ProxyStore"""
import logging
import warnings
from typing import Any, Dict, Union, List, Optional

import proxystore
from proxystore.proxy import extract
from proxystore.proxy import is_resolved
from proxystore.proxy import Proxy
from proxystore.store.base import Store
from proxystore.store.utils import resolve_async, get_key

proxystore_version = tuple(int(v) for v in proxystore.__version__.split('.'))
if proxystore_version >= (0, 7, 0):
    # In ProxyStore v0.7 and later, the Store config is a pydantic model.
    from proxystore.store.config import StoreConfig
else:
    # In ProxyStore v0.6 and older, the Store config is a dictionary.
    StoreConfig = Dict

logger = logging.getLogger(__name__)


class ProxyJSONSerializationWarning(Warning):
    pass


def get_store(
    name: str,
    config: Optional[Dict[str, Any]] = None,
) -> Optional[Store]:
    """Get a Store by name or create one if it does not already exist.

    Args:
        name (str): name of the store.
        config: ``Store`` configuration that can be used to reinitialize the
            ``Store`` if provided and a store with `name` is not found.

    Returns:
        The store registered as `name` or a newly intialized and registered
        store if `kind` is not ``None``.
    """
    store = proxystore.store.get_store(name)
    if store is None and config is not None:
        store = Store.from_config(config)
        # If the config has `register=True`, then the above line will
        # register the store for us. So we set `exist_ok=True` so the next
        # line does not raise an error. This only applies for
        # ProxyStore v0.6.5 and later.
        proxystore.store.register_store(store, exist_ok=True)
    return store


def proxy_json_encoder(proxy: Proxy) -> Any:
    """Custom encoder function for proxies

    Proxy objects are not JSON serializable so this function, when passed to
    `json.dumps()`, will attempt to JSON serialize the wrapped object. If the
    proxy is not resolved, a warning will be raised for the user and the
    proxy will be replaced with a placeholder string for the proxy. This
    1) prevents JSON serialization from failing and 2) avoid unintended
    resolutions of proxies that may invoke expensive communication operations
    without the user being aware.

    Usage:
        >>> # With JSON dump/dumps
        >>> json.dumps(json_obj_containing_proxy, default=proxy_json_encoder)
        >>> # With Pydantic
        >>> my_basemodel_instance.json(encoder=proxy_json_encoder)

    Args:
        proxy (Proxy): proxy to convert to JSON encodable object

    Returns:
        The object wrapped by the proxy if the proxy has already been resolved
        otherwise a placeholder string.

    Raises:
        TypeError:
            if `proxy` is not an instance of a Proxy.
    """
    if not isinstance(proxy, Proxy):
        # The JSON encoder will catch this TypeError and handle appropriately
        logger.error(f'Passed a series of objects that are not serializable: {type(proxy)}. {proxy}')
        raise TypeError(f'Unserializable type: {type(proxy)}')

    if is_resolved(proxy):
        # Proxy is already resolved so encode the underlying object
        # rather than the proxy
        return extract(proxy)

    warnings.warn(
        "Attemping to JSON serialize an unresolved proxy. To prevent "
        "an unintended proxy resolve, the resulting JSON object will "
        "have unresolved proxies replaced with a placeholder string.",
        category=ProxyJSONSerializationWarning
    )
    return f"<Unresolved Proxy at {hex(id(proxy))}>"


def resolve_proxies_async(args: Union[object, list, tuple, dict]) -> List[Proxy]:
    """Begin asynchronously resolving all proxies in input

    Scan inputs for instances of `Proxy` and begin asynchronously resolving.
    This is useful if you have one or more proxies that will be needed soon
    so the underlying objects can be asynchronously resolved to reduce the
    cost of the first access to the proxy.

    Args:
        args (object, list, tuple, dict): possible object or
            iterable of objects that may be ObjectProxy instances

    Returns:
        List of the proxies that are being resolved
    """

    # Create a list to store the keys
    output = []

    # Make a function that will resolve proxies
    def resolve_async_if_proxy(obj: Any) -> None:
        if isinstance(obj, Proxy):
            output.append(obj)
            resolve_async(obj)

    if isinstance(args, Proxy):
        resolve_async_if_proxy(args)
    elif isinstance(args, list) or isinstance(args, tuple):
        for x in args:
            resolve_async_if_proxy(x)
    elif isinstance(args, dict):
        for x in args:
            resolve_async_if_proxy(args[x])
    return output


def store_proxy_stats(proxy: Proxy, proxy_timing: dict):
    """Store the timings associated with a proxy, if available

    Args:
        proxy: Proxy to evaluate
        proxy_timing: Dictionary in which to store timings to be updated
    """
    # Get the key associated with this proxy
    key = get_key(proxy)

    # ProxyStore keys are NamedTuples, so we cast to a string to use as a JSON key.
    key = str(key)

    # Get the store associated with this proxy
    store = proxystore.store.get_store(proxy)
    if store.metrics is not None:
        # Get the stats and convert them to a JSON-serializable form
        metrics = store.metrics.get_metrics(proxy)
        stats = metrics.as_dict() if metrics is not None else {}
    else:
        stats = {}

    # Update existing timings
    if key not in proxy_timing:
        proxy_timing[key] = {}
    proxy_timing[key].update(stats)



"""Use Python's :class:`~concurrent.futures.Executor` to run workers on a local system"""
from typing import Optional, Union, Callable, Collection
from concurrent.futures import Future, Executor, ThreadPoolExecutor, ProcessPoolExecutor

from colmena.models.results import Result

from .base import FutureBasedTaskServer, convert_to_colmena_method
from colmena.queue.base import ColmenaQueues
from colmena.models.methods import ColmenaMethod


class LocalTaskServer(FutureBasedTaskServer):
    """Use Python's native concurrent libraries to execute tasks"""

    def __init__(self,
                 queues: ColmenaQueues,
                 methods: Collection[Union[Callable, ColmenaMethod]],
                 threads: bool = True,
                 num_workers: Optional[int] = None):
        """
        Args:
            methods: Methods to be served
            queues: Queues used to communicate with thinker
            threads: Use threads instead of workers
            num_workers: Number of workers to deploy.
        """
        self._methods = dict(
            (m.name, m) for m in map(convert_to_colmena_method, methods)
        )
        self._executor: Optional[Executor] = None
        self.num_workers = num_workers
        self.threads = threads
        super().__init__(queues=queues, method_names=list(self._methods.keys()))

    def _submit(self, task: Result, topic: str) -> Optional[Future]:
        return self._executor.submit(self._methods[task.method], task)

    def _setup(self):
        self._executor = (
            ThreadPoolExecutor(self.num_workers)
            if self.threads else
            ProcessPoolExecutor(self.num_workers)
        )

    def _cleanup(self):
        self._executor.shutdown()



from concurrent.futures import Future
from time import sleep
from uuid import uuid4

from pytest_mock import MockFixture
from pytest import fixture, mark

from colmena.queue.python import PipeQueues
from colmena.task_server.globus import GlobusComputeTaskServer

my_funcs: dict = {}


class FakeClient:
    """Faked Client that allows you to register functions"""

    def register_function(self, new_func, function_name: str = None, **kwargs):
        global my_funcs
        uuid = uuid4()
        if function_name is None:
            function_name = new_func.__name__
        my_funcs[uuid] = (new_func, function_name)
        return uuid


# TODO (wardlt): This mocked version does not mimic the real versions' multi-threaded nature, which
#  means it does not suffer the same problems when being copied over to a subprocess
class FakeExecutor:
    """Faked FuncXExecutor that generates "futures" but does not communicate with FuncX"""

    def __init__(self, *args, **kwargs):
        pass

    def submit_to_registered_function(self, func: str, kwargs: dict):
        # Get the function from the global registry
        global my_funcs
        func = my_funcs[func][0]
        task = kwargs['result']
        new_future = Future()
        result = func(task)
        new_future.set_result(result)
        return new_future

    def shutdown(self, *args, **kwargs):
        pass


@fixture()
def mock_globus(mocker: MockFixture):
    mocker.patch('colmena.task_server.globus.Executor', FakeExecutor)


@mark.timeout(10)
def test_mocked_server(mock_globus):
    # Create the task server with a single, no-op function
    client = FakeClient()
    queues = PipeQueues()

    def func(x):
        if x is None:
            raise MemoryError()
        return x

    fts = GlobusComputeTaskServer({func: 'fake_endp'}, client, queues)
    fts.start()

    # Submit a task to the queue and see how it works
    try:
        # Send a task that will execute properly
        queues.send_inputs(1, method='func')
        sleep(1)
        result = queues.get_result()
        assert result.success
        assert result.value == 1

        # Send a task that will throw a memory error
        queues.send_inputs(None, method='func')
        sleep(1)
        result = queues.get_result()
        assert not result.success
        assert 'MemoryError' in result.failure_info.exception

    finally:
        queues.send_kill_signal()
        fts.join()



"""Data models for requests for and results from compuations"""
import json
import logging
import sys
from math import nan
import pickle as pkl
from datetime import datetime
from enum import Enum
from functools import partial
from time import perf_counter
from traceback import TracebackException
from typing import Any, Tuple, Dict, Optional, Union, List, Sequence
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field
from proxystore.proxy import Proxy

from colmena.proxy import get_store, store_proxy_stats, StoreConfig, proxy_json_encoder

logger = logging.getLogger(__name__)


class SerializationMethod(str, Enum):
    """Serialization options"""

    JSON = "json"  # Serialize using JSONf
    PICKLE = "pickle"  # Pickle serialization

    @staticmethod
    def serialize(method: 'SerializationMethod', data: Any) -> str:
        """Serialize an object using a specified method

        Args:
            method: Method used to serialize the object
            data: Object to be serialized
        Returns:
            Serialized data
        """

        if method == "json":
            return json.dumps(data)
        elif method == "pickle":
            return pkl.dumps(data).hex()
        else:
            raise NotImplementedError(f'Method {method} not yet implemented')

    @staticmethod
    def deserialize(method: 'SerializationMethod', message: str) -> Any:
        """Deserialize an object

        Args:
            method: Method used to serialize
            message: Message to deserialize
        Returns:
            Result object
        """

        if method == "json":
            return json.loads(message)
        elif method == "pickle":
            return pkl.loads(bytes.fromhex(message))
        else:
            raise NotImplementedError(f'Method {method} not yet implemented')


def _serialized_str_to_bytes_shim(
        s: str,
        method: Union[str, SerializationMethod],
) -> bytes:
    """Shim between Colmena serialized objects and bytes.

    Colmena's serialization mechanisms produce strings but ProxyStore
    serializes to bytes, so this shim takes an object serialized by Colmena
    and converts it to bytes.

    Args:
        s: Serialized string object
        method: Serialization method used to produce s

    Returns:
        bytes representation of s
    """
    if method == "json":
        return s.encode('utf-8')
    elif method == "pickle":
        # In this case the conversion goes from obj > bytes > str > bytes
        # which results in an unnecessary conversion to a string but this is
        # an unavoidable side effect of converting between the Colmena
        # and ProxyStore serialization formats.
        return bytes.fromhex(s)
    else:
        raise NotImplementedError(f'Method {method} not yet implemented')


def _serialized_bytes_to_obj_wrapper(
        b: str,
        method: Union[str, SerializationMethod],
) -> Any:
    """Wrapper which converts bytes to strings before deserializing.

    Args:
        b: Byte string of serialized object
        method: Serialization method used to produce b

    Returns:
        Deserialized object
    """
    if method == "json":
        s = b.decode('utf-8')
    elif method == "pickle":
        # In this case the conversion goes from bytes > str > bytes > obj
        # which results in an unecessary conversion to a string but this is
        # an unavoidable side effect of converting between the Colmena
        # and ProxyStore serialization formats.
        s = b.hex()
    else:
        raise NotImplementedError(f'Method {method} not yet implemented')

    return SerializationMethod.deserialize(method, s)


class FailureInformation(BaseModel):
    """Stores information about a task failure"""

    exception: str = Field(..., description="The exception returned by the failed task")
    traceback: Optional[str] = Field(None, description="Full stack trace for exception, if available")

    @classmethod
    def from_exception(cls, exc: BaseException) -> 'FailureInformation':
        tb = TracebackException.from_exception(exc)
        return cls(exception=repr(exc), traceback="".join(tb.format()))


class WorkerInformation(BaseModel):
    """Information about the worker that executed this task"""

    hostname: Optional[str] = Field(None, description='Hostname of the worker who executed this task')

    model_config = ConfigDict(extra='allow')


class ResourceRequirements(BaseModel):
    """Resource requirements for tasks. Used by some Colmena backends to allocate resources to the task

    Follows the naming conventions of
    `RADICAL-Pilot <https://radicalpilot.readthedocs.io/en/stable/apidoc.html#taskdescription>`_.
    """

    # Defining how we use CPU resources
    node_count: int = Field(1, description='Total number of nodes to use for the task')
    cpu_processes: int = Field(1, description='Total number of MPI ranks per node')
    cpu_threads: int = Field(1, description='Number of threads per process')

    @property
    def total_ranks(self) -> int:
        """Total number of MPI ranks"""
        return self.node_count * self.cpu_processes


class Timestamps(BaseModel):
    """A class which records the system times at which key events in a task occurred

    All should be in UTC.
    """

    created: float = Field(description="Time this value object was created",
                           default_factory=lambda: datetime.now().timestamp())
    input_received: float = Field(nan, description="Time the inputs was received by the task server")
    compute_started: float = Field(nan, description="Time workflow process began executing a task")
    compute_ended: float = Field(nan, description="Time workflow process finished executing a task")
    result_sent: float = Field(nan, description="Time message was sent from the server")
    result_received: float = Field(nan, description="Time value was received by client")
    start_task_submission: float = Field(nan, description="Time marking the start of the task submission to workflow engine")
    task_received: float = Field(nan, description="Time task result received from workflow engine")


class TimeSpans(BaseModel):
    """Amount of time elapsed between major events

    All are recorded in seconds
    """

    running: float = Field(nan, description="Runtime of the method, if available")
    serialize_inputs: float = Field(nan, description="Time required to serialize inputs on client")
    deserialize_inputs: float = Field(nan, description="Time required to deserialize inputs on worker")
    serialize_results: float = Field(nan, description="Time required to serialize results on worker")
    deserialize_results: float = Field(nan, description="Time required to deserialize results on client")
    async_resolve_proxies: float = Field(nan, description="Time required to start async resolves of proxies")
    proxy: Dict[str, Dict[str, dict]] = Field(default_factory=dict,
                                              description='Timings related to resolving ProxyStore proxies on the compute worker')

    additional: Dict[str, float] = Field(default_factory=dict,
                                         description="Additional timings reported by a task server")


class Result(BaseModel):
    """A class which describes the inputs and results of the calculations evaluated by the MethodServer

    Each instance of this class stores the inputs and outputs to the function along with some tracking
    information allowing for performance analysis (e.g., time submitted to Queue, time received by client).
    All times are listed as Unix timestamps.

    The Result class also handles serialization of the data to be transmitted over a RedisQueue
    """

    # Core result information
    task_id: str = Field(default_factory=lambda: str(uuid4()), description='Unique identifier for each task')
    inputs: Union[Tuple[Tuple[Any, ...], Dict[str, Any]], str] = \
        Field(None, description="Input to a function. Positional and keyword arguments. The `str` data type "
                                "is for internal use and is used when communicating serialized objects.")
    value: Any = Field(None, description="Output of a function")
    method: Optional[str] = Field(None, description="Name of the method to run.")
    success: Optional[bool] = Field(None, description="Whether the task completed successfully")
    complete: Optional[bool] = Field(None, description="Whether this result is the last for a task instead of an intermediate result")

    # Store task information
    task_info: Optional[Dict[str, Any]] = Field(default_factory=dict,
                                                description="Task tracking information to be transmitted "
                                                            "along with inputs and results. User provided")
    resources: ResourceRequirements = Field(default_factory=ResourceRequirements, description='List of the resources required for a task, if desired')
    failure_info: Optional[FailureInformation] = Field(None, description="Messages about task failure. Provided by Task Server")
    worker_info: Optional[WorkerInformation] = Field(None, description="Information about the worker which executed a task. Provided by Task Server")
    message_sizes: Dict[str, int] = Field(default_factory=dict, description='Sizes of the inputs and results in bytes')

    # Timings
    timestamp: Timestamps = Field(default_factory=Timestamps, description='Times at which major events occurred')
    time: TimeSpans = Field(default_factory=TimeSpans, description='Elapsed time between major events')

    # Serialization options
    serialization_method: SerializationMethod = Field(SerializationMethod.JSON,
                                                      description="Method used to serialize input data")
    keep_inputs: bool = Field(True, description="Whether to keep the inputs with the result object or delete "
                                                "them after the method has completed")
    proxystore_name: Optional[str] = Field(None, description="Name of ProxyStore backend you use for transferring large objects")
    proxystore_config: Optional[StoreConfig] = Field(None, description="ProxyStore backend configuration")
    proxystore_threshold: Optional[int] = Field(None,
                                                description="Proxy all input/output objects larger than this threshold in bytes")

    # Task routing information
    topic: Optional[str] = Field(None, description='Label used to group results in queue between Thinker and Task Server')

    # Fault tolerance
    max_retries: int = Field(0, description='Maximum number of times this task should be retried if it fails')
    retries: int = Field(0, description='Number of times this task has been retried')

    def __init__(self, inputs: Tuple[Tuple[Any], Dict[str, Any]], **kwargs):
        """
        Args:
             inputs (Any, Dict): Inputs to a function. Separated into positional and keyword arguments
        """
        super().__init__(inputs=inputs, **kwargs)

    @classmethod
    def from_args_and_kwargs(cls, fn_args: Sequence[Any], fn_kwargs: Dict[str, Any] = None, **kwargs):
        """Create a result object form a the arguments and kwargs for the function

        Keyword arguments to this function are passed to the initializer for `Result`.

        Args:
            fn_args: Positional arguments to the function
            fn_kwargs: Keyword arguments to the function
        Returns:
            Result object with the results object
        """
        return cls(inputs=(tuple(fn_args), fn_kwargs or dict()), **kwargs)

    @property
    def args(self) -> Tuple[Any]:
        return tuple(self.inputs[0])

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self.inputs[1]

    def json(self, **kwargs: Dict[str, Any]) -> str:
        """Override json encoder to use a custom encoder with proxy support"""
        if 'exclude' in kwargs:
            # Make a shallow copy of the user passed excludes
            user_exclude = kwargs['exclude'].copy()
            if isinstance(kwargs['exclude'], dict):
                kwargs['exclude'].update({'inputs': True, 'value': True})
            if isinstance(kwargs['exclude'], set):
                kwargs['exclude'].update({'inputs', 'value'})
            else:
                raise ValueError(
                    f'Unsupported type {type(kwargs["exclude"])} for argument "exclude". Expected set or dict')
        else:
            user_exclude = set()
            kwargs['exclude'] = {'inputs', 'value'}

        # Use pydantic's encoding for everything except `inputs` and `values`
        data = self.model_dump(**kwargs)

        # Add inputs/values back to data unless the user excluded them
        if isinstance(user_exclude, set):
            if 'inputs' not in user_exclude:
                data['inputs'] = self.inputs
            if 'value' not in user_exclude:
                data['value'] = self.value
        elif isinstance(user_exclude, dict):
            if not user_exclude['inputs']:
                data['inputs'] = self.inputs
            if not user_exclude['value']:
                data['value'] = self.value

        # Jsonify with custom proxy encoder
        return json.dumps(data, default=proxy_json_encoder)

    def mark_result_received(self):
        """Mark that a completed computation was received by a client"""
        self.timestamp.result_received = datetime.now().timestamp()

    def mark_input_received(self):
        """Mark that a task server has received a value"""
        self.timestamp.input_received = datetime.now().timestamp()

    def mark_compute_started(self):
        """Mark that the compute for a method has started"""
        self.timestamp.compute_started = datetime.now().timestamp()

    def mark_result_sent(self):
        """Mark when a result is sent from the task server"""
        self.timestamp.result_sent = datetime.now().timestamp()

    def mark_start_task_submission(self):
        """Mark when the Task Server submits a task to the engine"""
        self.timestamp.start_task_submission = datetime.now().timestamp()

    def mark_task_received(self):
        """Mark when the Task Server receives the task from the engine"""
        self.timestamp.task_received = datetime.now().timestamp()

    def mark_compute_ended(self):
        """Mark when the task finished executing"""
        self.timestamp.compute_ended = datetime.now().timestamp()

    def set_result(self, result: Any, runtime: float = nan, intermediate: bool = False):
        """Set the value of this computation

        Automatically sets the "time_result_completed" field and, if known, defines the runtime.

        Will delete the inputs to the function if the user specifies ``self.return_inputs == False``.
        Removing the inputs once the result is known can save communication time

        Args:
            result: Result to be stored
            runtime: Runtime for the function
            intermediate: If this result is not the final one in a workflow
        """
        self.value = result
        if not self.keep_inputs:
            self.inputs = ((), {})
        self.time.running = runtime
        self.success = True
        self.complete = not intermediate

    def serialize(self) -> Tuple[float, List[Proxy]]:
        """Stores the input and value fields as a pickled objects

        Returns:
            - (float) Time to serialize
            - List of any proxies that were created
        """
        start_time = perf_counter()
        _value = self.value
        _inputs = self.inputs
        proxies = []

        if self.proxystore_name is not None:
            store = get_store(name=self.proxystore_name, config=self.proxystore_config)
        else:
            store = None

        def _serialize_and_proxy(value, evict=False) -> Tuple[str, int]:
            """Helper function for serializing and proxying

            Args:
                value: Value to be serialized
                evict: Whether to evict from proxy store on reading
            Returns:
                - Serialized representation, which is either the object or a proxy to it
                - Size of the serialized object (not the size of the proxy)
            """
            # Serialized object before proxying to compare size of serialized
            # object to value server threshold. Using sys.getsizeof would be
            # faster but sys.getsizeof does not account for the memory
            # consumption of objects that value refers to
            value_str = SerializationMethod.serialize(
                self.serialization_method, value
            )
            value_size = sys.getsizeof(value_str)

            if (
                    store is not None and
                    self.proxystore_threshold is not None and
                    not isinstance(value, Proxy) and
                    value_size >= self.proxystore_threshold
            ):
                # Override ProxyStore's default serialization with these shims
                # to Colmena's serialization mechanisms. This avoids value
                # being serialized twice: once to get the size of the
                # serialized object and once by proxy().
                deserializer = partial(
                    _serialized_bytes_to_obj_wrapper,
                    method=self.serialization_method,
                )
                serializer = partial(
                    _serialized_str_to_bytes_shim,
                    method=self.serialization_method,
                )

                value_proxy = store.proxy(
                    value_str,
                    evict=evict,
                    deserializer=deserializer,
                    serializer=serializer,
                )
                logger.debug(f'Proxied object of type {type(value)} with id={id(value)}')
                proxies.append(value_proxy)

                # Update the statistics
                store_proxy_stats(value_proxy, self.time.proxy)

                # Serialize the proxy with Colmena's utilities. This is
                # efficient since the proxy is just a reference and metadata
                value_str = SerializationMethod.serialize(
                    self.serialization_method, value_proxy
                )

            return value_str, value_size

        try:
            # Each value in *args and **kwargs is serialized independently
            if len(_inputs[0]) > 0:
                args, args_sizes = zip(*map(_serialize_and_proxy, _inputs[0]))
            else:
                args = args_sizes = []
            kwargs = {}
            kwarg_sizes = []
            for k, v in _inputs[1].items():
                _kwarg_str, _kwarg_size = _serialize_and_proxy(v)
                kwargs[k] = _kwarg_str
                kwarg_sizes.append(_kwarg_size)
            self.inputs = (args, kwargs)

            # Store the size of the input if not already there
            if 'inputs' not in self.message_sizes:
                self.message_sizes['inputs'] = sum(args_sizes) + sum(kwarg_sizes)

            # The entire result is serialized as one object. Pass evict=True
            # so the value is evicted from the value server once it is resolved
            # by the thinker.
            if _value is not None:
                self.value, size = _serialize_and_proxy(_value, evict=True)
                if 'value' not in self.message_sizes:
                    self.message_sizes['value'] = size

            return perf_counter() - start_time, proxies
        except Exception as e:
            # Put the original values back
            self.inputs = _inputs
            self.value = _value
            raise e

    def deserialize(self) -> float:
        """De-serialize the input and value fields

        Returns:
            (float) The time required to deserialize
        """
        # Check that the data is actually a string
        start_time = perf_counter()
        _value = self.value
        _inputs = self.inputs

        def _deserialize(value):
            if not isinstance(value, str):
                return value
            return SerializationMethod.deserialize(self.serialization_method, value)

        if isinstance(_inputs, str):
            _inputs = SerializationMethod.deserialize(self.serialization_method, _inputs)

        try:
            # Deserialize each value in *args and **kwargs
            args = tuple(map(_deserialize, _inputs[0]))
            kwargs = {k: _deserialize(v) for k, v in _inputs[1].items()}
            self.inputs = (args, kwargs)

            # Deserialize result if it exists
            if _value is not None:
                self.value = _deserialize(_value)

            return perf_counter() - start_time
        except Exception as e:
            # Put the original values back
            self.inputs = _inputs
            self.value = _value
            raise e



"""Parsl task server and related utilities"""
import os
import shlex
import logging
import platform
import shutil
from concurrent.futures import Future
from functools import partial
from pathlib import Path
from tempfile import mkdtemp
from time import perf_counter
from datetime import datetime
from typing import Optional, List, Callable, Tuple, Dict, Union

import parsl
from parsl.app.app import AppBase
from parsl.app.bash import BashApp
from parsl.config import Config
from parsl.app.python import PythonApp
from colmena.models.methods import ExecutableMethod

from colmena.queue.base import ColmenaQueues
from colmena.models import Result, FailureInformation, ResourceRequirements
from colmena.proxy import resolve_proxies_async
from colmena.task_server.base import convert_to_colmena_method, FutureBasedTaskServer

logger = logging.getLogger(__name__)


# Functions related to splitting "ExecutableTasks" into multiple steps
def _execute_preprocess(task: ExecutableMethod, result: Result) -> Tuple[Result, Path, Tuple[List[str], Optional[str]]]:
    """Perform the pre-processing step for an executable task

    Must execute on the remote system

    Manages pulling inputs down from the ProxyStore,
     creating a temporary run directory,
     and calling ExecutableTask.preprocess to store
     any needed items in that run directory

    Args:
        task: Description of task to be executed
        result: Object holding the inputs
    Returns:
        - Updated result object
        - Path to the temporary directory
        - A tuple of the inputs to the executable
            - List of CLI arguments
            - Contents of stdin, if desired
    """

    # Mark that compute has started
    result.mark_compute_started()

    # Unpack the inputs
    result.time.deserialize_inputs = result.deserialize()

    # Start resolving any proxies in the input asynchronously
    start_time = perf_counter()
    resolve_proxies_async(result.args)
    resolve_proxies_async(result.kwargs)
    result.time.async_resolve_proxies = perf_counter() - start_time

    # Create a temporary directory
    #  TODO (wardlt): Figure out how to allow users to define a path for temporary directories
    temp_dir = Path(mkdtemp(prefix='colmena_'))

    # Execute the function
    start_time = perf_counter()
    try:
        output = task.preprocess(temp_dir, result.args, result.kwargs)
    except BaseException as e:
        output = ([], None)  # Set a null value that still matches the output type
        result.success = False
        result.failure_info = FailureInformation.from_exception(e)
    finally:
        end_time = perf_counter()

    # Record the time required to perform the pre-processing
    result.time.additional['exec_preprocess'] = end_time - start_time

    # Remove the inputs. We don't need to send them back to the manager (the manager already knows what it sent out)
    result.inputs = ((), {})

    return result, temp_dir, output


def _execute_postprocess(task: ExecutableMethod, exit_code: int, result: Result, temp_dir: Path, serialized_inputs: str):
    """Execute the post-processing function after an executable completes

    Args:
        task: Task description, which contains details on how to post-process the results
        exit_code: Exit code the application (should be 0)
        result: Storage for the result data
        temp_dir: Path to the run directory on the remote system
        serialized_inputs: Copy of the serialized inputs
            The ``Result`` object is currently without a copy of the inputs
    """

    # Execute the function
    start_time = perf_counter()
    try:
        output = task.postprocess(temp_dir)
        result.success = True
        shutil.rmtree(temp_dir)
    except BaseException as e:
        output = None
        result.success = False
        result.failure_info = FailureInformation.from_exception(e)
    finally:
        end_time = perf_counter()
    result.time.additional['exec_postprocess'] = end_time - start_time

    # Store the results
    if result.success:
        result.set_result(output, datetime.now().timestamp() - result.timestamp.compute_started)

    # Store the run time in the result object
    result.time.additional['exec_execution'] = (result.time.running -
                                                result.time.additional['exec_postprocess'] -
                                                result.time.additional['exec_preprocess'])

    # Add the worker information into the tasks, if available
    worker_info = {'hostname': platform.node()}
    result.worker_info = worker_info

    # Re-pack the results (will use proxystore, if able)
    result.time.serialize_results, _ = result.serialize()

    # Put the serialized inputs back, if desired
    if result.keep_inputs:
        result.inputs = serialized_inputs

    return result


def _execute_execute(task: ExecutableMethod, task_path: Path, arguments: List[str],
                     stdin: Optional[str], cpu_process_type: str, *,
                     stdout: str, stderr: str, pre_exec: str = None, **kwargs) -> str:
    """Execute the executable step of an executable task

    This function is executed after :meth:`__execute_preprocess` has completed, which means
    any necessary input files are in `task_path` and any necessary CLI arguments have been determined.

    It will be executed as a :class:`BashApp`, so it returns a run command given the arguments provided.

    The kwargs for the argument include resource requirements and other information used to communicate
    the task requirements to the Parsl Executor.
    The primary user of the such kwargs is the RCT backend for Parsl.

    Args:
        task: General task information. Includes the path to the executable
        task_path: Path to the run directory
        arguments: List of arguments to add to the function execution
        stdin: Data to be passed to the stdin (not currently supported)
        cpu_process_type: Process type "MPI" or "SINGLE." Used by RCT to determine how to execute the program
        pre_exec: List of environment variables to set
        stdout: Path to the stdout for the function (should be ``task_task_path // 'colmena.stdout`).
            Provided as a kwargs so that the Parsl executor knows where to write the file
        stderr: Path to the stderr for the function (should be ``task_task_path // 'colmena.stdout`).
            Provided as a kwargs so that the Parsl executor knows where to write the file
        kwargs: Extra keyword arguments define the resource requirements. See :class:`ResourceRequirements`.
    Returns:
        The function to invoke as a string
    """

    assert stdin is None or len(stdin) == 0, "Standard in is not supported yet"

    # Make the launch command
    resources = ResourceRequirements.parse_obj(kwargs)
    shell_cmd = task.assemble_shell_cmd(arguments, resources)

    # Create the shell command
    #  TODO (wardlt): This is shlex.join, which is only available in Py3.8+
    shell_cmd = " ".join(shlex.quote(str(s)) for s in shell_cmd)

    # Move to the run directory
    os.chdir(task_path)

    return shell_cmd


def _preprocess_callback(
        preprocess_future: Future,
        serialized_inputs: str,
        result: Result,
        task_server: 'ParslTaskServer',
        topic: str,
        execute_fun: AppBase,
        postprocess_fun: AppBase
):
    """Perform the next steps in an executable workflow

    If preprocessing was unsuccessful, send failed result back to client.
    If successful, submit the "execute task" to Parsl and attach a callback
    to that function which will submit the postprocess task.

    Args:
        preprocess_future: Future provided when submitting pre-process task to Parsl
        serialized_inputs: Original inputs, still in serialized form.
            We deserialize the inputs in `_execute_preprocess` and do not pass the input data back to this callback function to minimize data sent.
        result: Result object to be gradually updated
        task_server: Connection to the Parsl task server. Used to send results back to client
        topic: Topic of the task
        execute_fun: Parsl App that submits the execution task
        postprocess_fun: Parsl App that performs post-processing
    """

    # If the Parsl execution was unsuccessful, send the result back to the client
    if preprocess_future.exception() is not None:
        return task_server.perform_callback(preprocess_future, result, topic)

    # If successful, unpack the outputs
    result, temp_dir, (exec_args, exec_stdin) = preprocess_future.result()

    # If unsuccessful, send the results back to the client
    if result.success is not None and not result.success:

        logger.info('Result failed during preprocessing. Sending back to client.')

        # Send the serialized inputs back to the client if they are expected
        if result.keep_inputs:
            result.inputs = serialized_inputs

        # Store the time it took to run the preprocessing
        result.time.running = result.time.additional.get('exec_preprocess', 0)
        return task_server.queues.send_result(result)

    # If successful, submit the execute step and pass its result to Parsl
    logger.info(f'Preprocessing was successful for {result.method} task. Submitting to execute')

    exec_future: Future = execute_fun(temp_dir, exec_args, exec_stdin,
                                      stdout=str(temp_dir / 'colmena.stdout'),
                                      stderr=str(temp_dir / 'colmena.stderr'),
                                      **result.resources.dict())

    # Submit post-process to collect the results of the exec_function
    post_future: Future = postprocess_fun(exec_future, result, temp_dir, serialized_inputs if result.keep_inputs else None)

    # Once that function completes, you are ready to submit the task back to the client
    def _send_back(future: Future):
        # Send the results back to the client, if desired
        #  This is only used if the task fails. (Otherwise, the copy of "result" held in the future is used)
        if result.keep_inputs:
            result.inputs = serialized_inputs
        return task_server.perform_callback(future, result, topic)

    post_future.add_done_callback(_send_back)


class ParslTaskServer(FutureBasedTaskServer):
    """Task server based on Parsl

    Create a Parsl task server by first creating a resource configuration following
    the recommendations in `the Parsl documentation
    <https://parsl.readthedocs.io/en/stable/userguide/configuring.html>`_.
    Then instantiate a task server with a list of Python functions,
    configurations defining on which Parsl executors each function can run,
    and the Parsl resource configuration.
    The executor(s) for each function can be defined with a combination
    of per method specifications

    .. code-block:: python

        ParslTaskServer([(f, {'executors': ['a']})], queue, config)

    and also using a default executor

    .. code-block:: python

        ParslTaskServer([f], queue, config, default_executors=['a'])

    Further configuration options for each method can be defined
    in the list of methods.

    **Technical Details**

    The task server stores each of the supplied methods as Parsl "Apps".
    Tasks are launched on remote workers by calling these Apps,
    and results are placed in the result queue by callbacks attached the resultant Parsl Futures.

    The behavior of an :class:`ExecutableTask` involves several Apps and callbacks.

    1. A :class:`PythonApp` to invoke the "preprocessing" function that is given the :class:`Result`.

       The app produces a path to a temporary run directory containing the input files,
       content for the standard input of the executable,
       and an updated copy of the :class:`Result` object containing timing information.

       Note that the :class:`Result` object returned by this app lacks the inputs to reduce communication costs.

       Once complete (successfully or not), it invokes a callback which launches the next two tasks
       and creates the next callback.
       In the even of an unsuccessful execution, the callback function returns the failure information to the client and exits.

    2. A :class:`BashApp` to run the executable that is given the path to the run directory
       and the list of resources required for executing the task.

       There is no callback for app.

    3. A :class:`PythonApp` to store the results of the execution that is given the exit code of the executable (should be 0),
       a copy of the :class:`Result` object produced by the preprocessing,
       the path to the run directory,
       and a serialized version of the inputs to the app.

       The application parses the outputs from the execution, stores them in the :class:`Result` object,
       and then serializes results for transmission back to the client.
       The application also re-inserts the inputs if they are required to be sent back to the client.

       The callback for this function submits the outputs, if successful, or any failure information, if not, to the result queue.

    Every one of the Apps is run on the remote system as they may involve manipulating files
    on the remote system.
    """

    def __init__(self, methods: List[Union[Callable, Tuple[Callable, Dict]]],
                 queues: ColmenaQueues,
                 config: Config,
                 timeout: Optional[int] = None,
                 default_executors: Union[str, List[str]] = 'all'):
        """

        Args:
            methods (list): List of methods to be served.
                Each element in the list is either a function or a tuple where the first element
                is a function and the second is a dictionary of the arguments being used to create
                the Parsl ParslApp see `Parsl documentation
                <https://parsl.readthedocs.io/en/stable/stubs/parsl.app.app.python_app.html#parsl.app.app.python_app>`_.
            queues: Queues for the task server
            config: Parsl configuration
            timeout (int): Timeout, if desired
            default_executors: Executor or list of executors to use by default.
        """
        # Get a list of default executors
        if default_executors == 'all':
            default_executors = [e.label for e in config.executors]

        # Store the Parsl configuration
        self.config = config

        # Assemble the list of methods
        self.methods_: Dict[str, Tuple[AppBase, str]] = {}  # Store the method and its type
        self.exec_apps_: Dict[str, Tuple[AppBase, AppBase]] = {}  # Stores the execute and post-process apps for ExecutableTasks
        for method in methods:
            # Get the options or use the defaults
            if isinstance(method, (tuple, list)):
                if len(method) != 2:
                    raise ValueError('Method description should a tuple of length 2')
                function, options = method
            else:
                function = method
                options = {'executors': default_executors}
                logger.info(f'Using default executors for {function.__name__}: {default_executors}')

            # Convert the function to a Colmena task
            function = convert_to_colmena_method(function)
            name = function.name

            # If the function is not an executable, submit it as a single task
            if not isinstance(function, ExecutableMethod):
                app = PythonApp(function, **options)
                self.methods_[name] = (app, 'basic')
            else:
                logger.info(f'Building a chain of apps for an ExecutableTask, {function.__name__}')
                # If it is an executable, the function we launch initially is a "preprocess inputs" function
                preprocess_fun = partial(_execute_preprocess, function)
                preprocess_fun.__name__ = f'{name}_preprocess'
                preprocess_app = PythonApp(preprocess_fun, **options)

                # Make executable app, which is just to invoke the executable
                execute_fun = partial(_execute_execute, function, cpu_process_type='MPI' if function.mpi else 'SINGLE')
                execute_fun.__name__ = f'{name}_execute'
                execute_app = BashApp(execute_fun, **options)

                # Make the post-process app, which gathers the results and puts them in the "result object"
                postprocess_fun = partial(_execute_postprocess, function)
                postprocess_fun.__name__ = f'{name}_postprocess'
                postprocess_app = PythonApp(postprocess_fun, **options)

                # Store them for use during submission phase
                self.methods_[name] = (preprocess_app, 'exec')
                self.exec_apps_[name] = (execute_app, postprocess_app)

        logger.info(f'Defined {len(self.methods_)} methods: {", ".join(self.methods_.keys())}')

        # Initialize the base class
        super().__init__(queues, self.methods_.keys(), timeout)

    def _submit(self, task: Result, topic: str) -> Optional[Future]:
        # Determine which method to run
        method = task.method

        # Submit the application
        task.mark_start_task_submission()
        function, func_type = self.methods_[method]
        serialized_inputs = task.inputs  # Hold a copy of the original inputs. Used for "exec" apps to minimize
        future: Future = function(task)
        logger.debug('Pushed task to Parsl')
        # TODO (wardlt): Implement "resubmit if task returns a new future." or the ability to launch Parsl workflows with >1 step

        # Depending on the task type, return a different future
        if func_type == 'basic':
            # For most functions, just return the future so the task server will handle the output
            return future
        elif func_type == 'exec':
            # For executable functions, we have a different route for returning results
            exec_app, post_app = self.exec_apps_[method]
            # TODO (wardlt): Use a join_app rather than callback?
            future.add_done_callback(lambda x: _preprocess_callback(x, serialized_inputs, task, self, topic, exec_app, post_app))
            return None  # `None` prevents the Task Server from adding its own callback
        else:
            raise ValueError(f'Unrecognized function type: {func_type}')

    def _cleanup(self):
        """Close out any resources needed by the task server"""
        # Wait until all tasks have finished
        dfk = parsl.dfk()
        dfk.wait_for_current_tasks()
        dfk.cleanup()
        logger.info(f"All tasks have completed for {self.__class__.__name__} on {self.ident}")

    def _setup(self):
        # Launch the Parsl workflow engine
        parsl.load(self.config)
        logger.info(f"Launched Parsl DFK. Process id: {os.getpid()}")
